from whad.dot15d4.stack.manager import Dot15d4Manager
from whad.dot15d4.stack.service import Dot15d4Service
from whad.dot15d4.stack.mac.constants import MACAddressMode, MACScanType, \
    MACDeviceType, MACPowerSource, MACConstants, MACAssociationStatus
from whad.dot15d4.stack.mac.network import Dot15d4PANNetwork

from whad.zigbee.stack.nwk.exceptions import NWKTimeoutException, NWKInvalidKey
from whad.zigbee.stack.nwk.database import NWKIB
from whad.zigbee.stack.nwk.neighbors import NWKNeighborTable
from whad.zigbee.stack.nwk.security import NetworkSecurityMaterial
from whad.zigbee.stack.nwk.network import ZigbeeNetwork
from whad.zigbee.stack.nwk.constants import ZigbeeDeviceType, NWKAddressMode, \
    NWKJoinMode, ZigbeeRelationship, BROADCAST_ADDRESSES

from whad.zigbee.stack.aps import APSManager

from whad.zigbee.crypto import NetworkLayerCryptoManager
from whad.common.stack import Layer, alias, source, state
from whad.exceptions import RequiredImplementation

from scapy.layers.zigbee import ZigBeeBeacon, ZigbeeNWKStub, ZigbeeNWK, \
    ZigbeeSecurityHeader, ZigbeeNWKCommandPayload, ZigbeeAppDataPayload, \
    ZigbeeAppDataPayloadStub, LinkStatusEntry
from whad.scapy.layers.zll import ZigbeeZLLCommissioningCluster, NewZigbeeAppDataPayloadStub

from random import randint
from scapy.fields import FlagValueIter

from time import time, sleep
import logging

logger = logging.getLogger(__name__)

class NWKService(Dot15d4Service):
    """
    This class represents a NWK service, exposing a standardized API.
    """
    def __init__(self, manager, name=None):
        super().__init__(
            manager,
            name=name,
            timeout_exception_class=NWKTimeoutException
        )


class NWKDataService(NWKService):
    """
    NWK service processing Data packets.
    """
    def __init__(self, manager):
        super().__init__(manager, name="nwk_data")

    @Dot15d4Service.request("NLDE-DATA")
    def data(
        self,
        nsdu,
        nsdu_handle=0,
        alias_address=None,
        alias_sequence_number=0,
        destination_address_mode=NWKAddressMode.UNICAST,
        destination_address=0xFFFF,
        radius=0,
        non_member_radius=7,
        discover_route=False,
        security_enable=False
    ):
        """
        Request allowing to transmit NSPDU to MAC layer.
        """
        if alias_address is not None:
            source_address = alias_address
            sequence_number = alias_sequence_number
        else:
            source_address = self.manager.get_layer('mac').get_service("management").get("macShortAddress")
            sequence_number = self.database.get("nwkSequenceNumber")
            self.database.set("nwkSequenceNumber", sequence_number+1)

        if radius == 0:
            radius = self.database.get("nwkMaxDepth") * 2

        flags = []
        if destination_address_mode == NWKAddressMode.MULTICAST:
            flags.append("multicast")
            raise RequiredImplementation("Routing/Multicasting")

        npdu = ZigbeeNWK(
            discover_route=int(discover_route),
            seqnum=sequence_number,
            radius=radius,
            flags=flags,
            destination=destination_address,
            source=source_address
        )
        if destination_address in BROADCAST_ADDRESSES.keys():
            destination_pan_id = self.database.get("nwkPANId")
            acknowledged = False
        else:
            destination_pan_id = self.database.get("nwkPANId")
            acknowledged = True

        if self.database.get("nwkSecurityLevel") != 0 and security_enable:
            npdu.flags.security = True
            selected_key_material = None

            security_material_set = self.database.get("nwkSecurityMaterialSet")
            for key_material in security_material_set:
                if key_material.key_sequence_number == self.database.get("nwkActiveKeySeqNumber"):
                    selected_key_material = key_material
                    break

            if selected_key_material is None:
                return False

            msdu = npdu / ZigbeeSecurityHeader(
                nwk_seclevel = self.database.get("nwkSecurityLevel"),
                source = self.database.get("nwkIeeeAddress"),
                key_type=1,
                key_seqnum=self.database.get("nwkActiveKeySeqNumber"),
                fc=selected_key_material.outgoing_frame_counter,
                data=bytes(nsdu)
            )
            crypto_manager = NetworkLayerCryptoManager(selected_key_material.key)
            msdu = crypto_manager.encrypt(msdu)
            selected_key_material.outgoing_frame_counter += 1
        else:
            msdu = npdu / nsdu

        return self.manager.get_layer('mac').get_service("data").data(
                    msdu,
                    msdu_handle=0,
                    source_address_mode=MACAddressMode.SHORT,
                    destination_pan_id=destination_pan_id,
                    destination_address=destination_address,
                    pan_id_suppressed=False,
                    sequence_number_suppressed=False,
                    wait_for_ack=acknowledged
        )


    def on_data_npdu(self, npdu, link_quality=255):
        """
        Callback processing NPDU from NWK manager.

        It forwards the NPDU to indicate_data method.
        """
        self.indicate_data(npdu, link_quality=link_quality)

    @Dot15d4Service.indication("NLDE-DATA")
    def indicate_data(self, npdu, link_quality=255):
        """
        Indication transmitting NPDU to upper layer.
        """
        destination_address_mode = (
            NWKAddressMode.MULTICAST if "multicast" in npdu.flags else
            NWKAddressMode.UNICAST
        )
        destination_address = npdu.destination
        source_address = npdu.source
        security_use = ZigbeeSecurityHeader in npdu and npdu[ZigbeeSecurityHeader].underlayer.__class__ is ZigbeeNWK
        if security_use:
            nsdu = ZigbeeAppDataPayload(npdu.data)
        else:
            nsdu = npdu[ZigbeeAppDataPayload]

        return (
            nsdu, {
                "destination_address_mode":destination_address_mode,
                "destination_address":destination_address,
                "source_address":source_address,
                "security_use":security_use,
                "link_quality":link_quality
            }
        )


class NWKManagementService(NWKService):
    """
    NWK service processing Management packets.
    """
    def __init__(self, manager):
        self.joining_thread_running = False
        self.joining_thread = None
        super().__init__(manager, name="nwk_management")


    @Dot15d4Service.request("NLME-RESET")
    def reset(self, warm_start=False):
        """
        Request resetting NWK layer to default state.
        """
        if warm_start:
            self.database.set("nwkNeighborTable",NWKNeighborTable())
            self.database.set("nwkRouteTable",[])
            return True
        else:
            confirm = self.manager.get_layer('mac').get_service("management").reset(set_default_pib=True)
            if confirm:
                self.database.reset()
                return True
        return False


    @Dot15d4Service.request("NLME-GET")
    def get(self, attribute):
        """
        Implement the NLME-GET request operation.

        Allows to get access to an attribute value stored in NWIB database.
        """
        return self.database.get(attribute)

    @Dot15d4Service.request("NLME-SET")
    def set(self, attribute, value):
        """
        Implement the NLME-SET request operation.

        Allows to set value to an attribute stored in NWIB database.
        """
        return self.database.set(attribute, value)


    @Dot15d4Service.request("NLME-ED-SCAN")
    def ed_scan(self, scan_channels=0x7fff800, scan_duration=2):
        """
        Implements the NLME-ED-SCAN request.

        Allows to perform an Energy Detection scan.
        """
        confirm = self.manager.get_layer('mac').get_service("management").scan(
            scan_type=MACScanType.ENERGY_DETECTION,
            channel_page=0,
            scan_channels=scan_channels,
            scan_duration=scan_duration
        )
        return confirm


    @Dot15d4Service.request("NLME-LEAVE")
    def leave(self, device_address=None, remove_children=False, rejoin=True):
        """
        Implements the NLME-LEAVE request.

        Allows to leave a ZigBee network.
        """
        network_address = self.database.get("nwkNetworkAddress")
        if network_address == 0xFFFF:
            # We are not present in the network, exit the procedure
            return False

        if device_address is None or device_address == self.database.get("nwkIeeeAddress"):
            parent = self.database.get("nwkNeighborTable").get_parent()
            if parent is None:
                return False

            sequence_number = self.database.get("nwkSequenceNumber")
            self.database.set("nwkSequenceNumber", sequence_number+1)

            # We are exiting the network, build the leave command
            leave_command = ZigbeeNWK(
                frametype=1,
                seqnum=sequence_number,
                destination=parent.address,
                source=network_address,
            )/ZigbeeNWKCommandPayload(
                cmd_identifier=4,
                remove_children=int(remove_children),
                request=1,
                rejoin=int(rejoin)
            )
            self.manager.get_layer('mac').get_service("data").data(
                leave_command,
                source_address_mode=MACAddressMode.SHORT,
                destination_pan_id=parent.pan_id,
                destination_address=parent.address,
                wait_for_ack=False
            )
            return True


    @Dot15d4Service.request("NLME-JOIN")
    def join(self, extended_pan_id, association_type=NWKJoinMode.NEW_JOIN, scan_channels=0x7fff800, scan_duration=4, join_as_router=False, rx_on_when_idle=True, mains_powered_device=False, security_enable=False, force=False):
        """
        Implements the NLME-JOIN request.

        Allows to join or rejoin a ZigBee network.
        """
        # In case of new join
        if association_type == NWKJoinMode.NEW_JOIN:
            # Try successively available neighbors that can be considered potential parents
            table = self.database.get("nwkNeighborTable")

            while True:
                candidate_parents = table.select_suitable_parent(extended_pan_id, self.database.get("nwkUpdateId"),
                                                                no_permit_check=force)
                if len(candidate_parents) == 0:
                    return False

                selected_parent = candidate_parents[0]
                for candidate_parent in candidate_parents[1:]:
                    if candidate_parent.depth < selected_parent.depth:
                        selected_parent = candidate_parent

                self.database.set("nwkParentInformation", 0)

                device_type = (
                    MACDeviceType.FFD if
                    join_as_router else
                    MACDeviceType.RFD
                )
                power_source = (
                    MACPowerSource.ALTERNATING_CURRENT_SOURCE if
                    mains_powered_device else
                    MACPowerSource.BATTERY_SOURCE
                )

                capability_information = (
                    0 |
                    (int(join_as_router) << 1) |
                    (int(mains_powered_device) << 2) |
                    (int(rx_on_when_idle) << 3) |
                    (0 << 4) |
                    (0 << 6) |
                    (1 << 7)
                )
                self.database.set("nwkCapabilityInformation", capability_information)
                # Trigger an association at the MAC layer
                if self.manager.get_layer('mac').get_service("management").associate(
                    channel_page=0,
                    channel=selected_parent.logical_channel,
                    coordinator_pan_id=selected_parent.pan_id,
                    coordinator_address=selected_parent.address,
                    device_type=device_type,
                    power_source=power_source,
                    idle_receiving=rx_on_when_idle,
                    allocate_address=True,
                    security_capability=False,
                    fast_association=False
                ):
                    logger.debug("[nwk_management_service] association is successful")
                    # Update database according to successful association
                    macShortAddress = self.manager.get_layer('mac').database.get("macShortAddress")
                    self.database.set("nwkNetworkAddress", macShortAddress)
                    self.database.set("nwkUpdateId", selected_parent.update_id)
                    self.database.set("nwkPANId", selected_parent.pan_id)
                    self.database.set("nwkExtendedPANID", extended_pan_id)
                    selected_parent.relationship = ZigbeeRelationship.IS_PARENT
                    if selected_parent.extended_address is not None and selected_parent.address is not None:
                        nwkAddressMap = self.database.get("nwkAddressMap")
                        nwkAddressMap[selected_parent.extended_address] = selected_parent.address
                    return True
                else:
                    logger.debug("[nwk_management_service] association is unsuccessful")
                    selected_parent.potential_parent = 0

        # In the case of a rejoin
        elif association_type == NWKJoinMode.REJOIN:
            # Check if we need to allocate a new address
            if self.database.get("nwkNetworkAddress") != 0xFFFF:
                network_address = self.database.get("nwkNetworkAddress")
                allocate_address = False
            else:
                network_address = randint(1,0xFFF0) if self.database.get("nwkNetworkAddress") == 0xFFFF else self.database.get("nwkNetworkAddress")
                allocate_address = False

            if self.database.get("nwkAddrAlloc") == 2:
                allocate_address = False

            # Perform a network discovery to infer surounding networks
            candidate_zigbee_networks  = self.network_discovery(
                scan_channels=scan_channels,
                scan_duration=scan_duration
            )
            selected_zigbee_network = None
            # Select the zigbee network according to provided extended PAN ID
            for candidate_zigbee_network in candidate_zigbee_networks:
                if candidate_zigbee_network.extended_pan_id == extended_pan_id:
                    selected_zigbee_network = candidate_zigbee_network
                    break

            if selected_zigbee_network is None:
                return False

            # Select the right channel accordingly
            self.manager.get_layer('mac').set_channel_page(0)
            self.manager.get_layer('mac').set_channel(selected_zigbee_network.channel)
            table = self.database.get("nwkNeighborTable")
            while True:
                candidate_parents = table.select_suitable_parent(
                    extended_pan_id,
                    self.database.get("nwkUpdateId"),
                    no_permit_check=True
                )
                if len(candidate_parents) == 0:
                    return False
                selected_parent = candidate_parents[0]
                for candidate_parent in candidate_parents[1:]:
                    if candidate_parent.depth < selected_parent.depth:
                        selected_parent = candidate_parent

                self.database.set("nwkParentInformation", 0)

                device_type = (
                                MACDeviceType.FFD if
                                join_as_router else
                                MACDeviceType.RFD
                )
                power_source = (
                    MACPowerSource.ALTERNATING_CURRENT_SOURCE if
                    mains_powered_device else
                    MACPowerSource.BATTERY_SOURCE
                )

                capability_information = (
                    0 |
                    (int(join_as_router) << 1) |
                    (int(mains_powered_device) << 2) |
                    (int(rx_on_when_idle) << 3) |
                    (0 << 4) |
                    (0 << 6) |
                    (1 << 7)
                )
                self.database.set("nwkCapabilityInformation", capability_information)

                radius = self.database.get("nwkMaxDepth") * 2

                sequence_number = self.database.get("nwkSequenceNumber")
                self.database.set("nwkSequenceNumber", sequence_number+1)

                # Build rejoin request
                npdu = ZigbeeNWK(
                    frametype=1,
                    discover_route=0,
                    seqnum=sequence_number,
                    radius=radius,
                    flags=["extended_src"], #, "extended_dst"],
                    destination=selected_parent.address,
                    source=network_address,
                    #ext_dst=selected_parent.extended_address, # ?
                    ext_src=self.database.get("nwkIeeeAddress")
                )
                nsdu = ZigbeeNWKCommandPayload(
                    cmd_identifier=6,
                    allocate_address=int(1),
                    security_capability=False,
                    receiver_on_when_idle=int(rx_on_when_idle),
                    power_source=int(power_source),
                    device_type=int(device_type),
                    alternate_pan_coordinator=0
                )


                if self.database.get("nwkSecurityLevel") != 0 and security_enable:
                    npdu.flags.security = True
                    selected_key_material = None

                    security_material_set = self.database.get("nwkSecurityMaterialSet")
                    for key_material in security_material_set:
                        if key_material.key_sequence_number == self.database.get("nwkActiveKeySeqNumber"):
                            selected_key_material = key_material
                            break

                    if selected_key_material is None:
                        return False

                    msdu = npdu / ZigbeeSecurityHeader(
                        nwk_seclevel = self.database.get("nwkSecurityLevel"),
                        source = self.database.get("nwkIeeeAddress"),
                        key_type=1,
                        key_seqnum=self.database.get("nwkActiveKeySeqNumber"),
                        fc=selected_key_material.outgoing_frame_counter,
                        data=bytes(nsdu)
                    )
                    crypto_manager = NetworkLayerCryptoManager(selected_key_material.key)
                    msdu = crypto_manager.encrypt(msdu)
                    selected_key_material.outgoing_frame_counter += 1
                else:
                    msdu = npdu / nsdu

                self.manager.get_layer('mac').get_service("data").data(
                    msdu,
                    destination_address_mode=MACAddressMode.SHORT,
                    source_address_mode=MACAddressMode.SHORT,
                    destination_pan_id=selected_parent.pan_id,
                    destination_address=selected_parent.address,
                    wait_for_ack=True
                )
                duration = (
                            self.manager.get_layer('mac').database.get("macResponseWaitTime") *
                            MACConstants.A_BASE_SUPERFRAME_DURATION *
                            self.manager.get_layer('phy').symbol_duration
                )
                self.manager.get_layer('mac').database.set("macPanId", selected_parent.pan_id)

                self.manager.get_layer('mac').get_service('management').poll(
                    coordinator_pan_id=selected_parent.pan_id,
                    coordinator_address=selected_parent.address
                )

                rejoin_response = None
                try:
                    rejoin_response = self.wait_for_packet(
                        lambda pkt:ZigbeeNWKCommandPayload in pkt and pkt.cmd_identifier==7,
                        timeout=3
                    )
                except NWKTimeoutException:
                    selected_parent.potential_parent = 0

                if rejoin_response is not None and rejoin_response.rejoin_status == 0:
                    self.database.set("nwkNetworkAddress", rejoin_response.network_address)
                    self.manager.get_layer('mac').database.set(
                        "macShortAddress",
                        rejoin_response.network_address
                    )
                    self.database.set("nwkUpdateId", selected_parent.update_id)
                    self.database.set("nwkPANId", selected_parent.pan_id)
                    self.database.set("nwkExtendedPANID", extended_pan_id)
                    selected_parent.relationship = ZigbeeRelationship.IS_PARENT
                    if selected_parent.extended_address is not None and selected_parent.address is not None:
                        nwkAddressMap = self.database.get("nwkAddressMap")
                        nwkAddressMap[selected_parent.extended_address] = selected_parent.address
                    return True
                else:
                    selected_parent.potential_parent = 0

    @Dot15d4Service.indication("NLME-JOIN")
    def indicate_join(self, network_address, extended_address, capability_information, rejoin=False, secure_rejoin=False):

        return (
            network_address,
            {
                "extended_address":extended_address,
                "capability_information":capability_information,
                "rejoin":rejoin,
                "secure_rejoin":secure_rejoin
            }
        )
    @Dot15d4Service.request("NLME-NETWORK-DISCOVERY")
    def network_discovery(self, scan_channels=0x7fff800, scan_duration=6):
        """
        Implements the NLME-NETWORK-DISCOVERY request.

        Allows to discover a network.
        """
        confirm = self.manager.get_layer('mac').get_service("management").scan(
            scan_type=MACScanType.ACTIVE,
            channel_page=0,
            scan_channels=scan_channels,
            scan_duration=scan_duration
        )

        zigbee_networks = []
        notifications_left = True
        while notifications_left:
            try:
                beacon = self.wait_for_packet(lambda pkt:ZigBeeBeacon in pkt, timeout=0.1)

                if beacon.pan_descriptor in confirm:
                    network = ZigbeeNetwork(beacon)
                    if network not in zigbee_networks:
                        zigbee_networks.append(network)

            except NWKTimeoutException:
                notifications_left = False
        return zigbee_networks

    @Dot15d4Service.request("NLME-PERMIT-JOINING")
    def permit_joining(self, duration=0xFF):
        self._stop_joining_timeout()
        if duration == 0:
            self.manager.get_layer("mac").database.set("macAssociationPermit", False)
            return True
        elif duration == 0xFF:
            self.manager.get_layer("mac").database.set("macAssociationPermit", True)
            return True
        else:
            self.manager.get_layer("mac").database.set("macAssociationPermit", True)
            self._start_joining_timeout(duration)
            return True

    def _stop_joining_timeout(self):
        self.joining_thread_running = False
        self.joining_thread = None

    def _start_joining_timeout(self, duration):
        self.joining_thread_running = True
        self.joining_thread = Thread(target=self._joining_timeout, args=(duration, ), daemon=True)
        self.joining_thread.start()

    def _joining_timeout(self, duration):
        starting_time = time()
        while self.joining_thread_running and (time() - starting_time) < duration:
            sleep(0.1)
        # Turn off joining timeout
        self.manager.get_layer("mac").database.set("macAssociationPermit", False)

    @Dot15d4Service.request("NLME-NETWORK-FORMATION")
    def network_formation(
                            self,
                            pan_id=None,
                            channel=None,
                            scan_channels=0x7fff800,
                            scan_duration=4,
                            beacon_order=15,
                            superframe_order=15,
                            battery_life_extension=False,
                            distributed_network=False,
                            distributed_network_address=0x0001
    ):
        """
        Implements the NLME-NETWORK-FORMATION request.
        """
        if distributed_network:
            raise RequiredImplementation("DistributedNetworkFormation")

        # Select the best channel according to an Energy Detection scan
        ed_reports = self.ed_scan(
            scan_channels=scan_channels,
            scan_duration=scan_duration
        )

        logger.info("Channel measurement")
        if channel is None:
            minimal_measurement = None
            for ed_report in ed_reports:
                if (
                    minimal_measurement is None or
                    minimal_measurement.max_sample > ed_report.max_sample
                ):
                    minimal_measurement = ed_report

            best_channel = minimal_measurement.channel_number
        else:
            best_channel = channel

        logger.info("Active scan")
        if pan_id is None:
            # Perform an active scan to prevent the use of already in use PAN ID
            active_scan_reports = self.network_discovery(
                scan_channels=scan_channels,
                scan_duration=scan_duration
            )

            # Pick a PAN ID which is not already in use
            active_networks_pan_ids = [network.dot15d4_pan_network.coord_pan_id for network in active_scan_reports]

            selected_pan_id = randint(1,0xFFFD)
            while selected_pan_id in active_networks_pan_ids:
                selected_pan_id = randint(1,0xFFFD)
        else:
            selected_pan_id = pan_id

        self.manager.get_layer('mac').set_channel_page(0)
        self.manager.get_layer('mac').set_channel(best_channel)

        # Select network address
        if not distributed_network:
            network_address = 0x0000
        else:
            network_address = distributed_network_address

        # Espressif hijack coord: self.database.set("nwkExtendedPANID", 0x6055f90000f714e4)
        # Update the mac layer with the selected address

        self.manager.get_layer('mac').set_short_address(network_address)
        self.database.set("nwkNetworkAddress", network_address)


        if self.database.get("nwkExtendedPANID") == 0:

            self.database.set(
                "nwkExtendedPANID",
                self.manager.get_layer('mac').database.get("macExtendedAddress")
            )

        self.database.set("nwkPANId", selected_pan_id)

        # Build ZigBee beacon payload
        beacon_payload = ZigBeeBeacon(
            proto_id = 0,
            stack_profile = self.database.get("nwkStackProfile"),
            nwkc_protocol_version = self.database.get("nwkcProtocolVersion"),
            router_capacity = 1, #((self.database.get("nwkCapabilityInformation") >> 1) & 1),
            end_device_capacity = 1,
            device_depth = 0 if network_address == 0 else 1,
            extended_pan_id = self.database.get("nwkExtendedPANID"),
            tx_offset = 0xFFFFFF, # non beacon network, default value
            update_id = self.database.get("nwkUpdateId")
        )
        self.manager.get_layer('mac').database.set("macBeaconPayload", bytes(beacon_payload))
        start = self.manager.get_layer('mac').get_service('management').start(
            selected_pan_id,
            pan_coordinator=True,
            beacon_order=beacon_order,
            superframe_order=superframe_order,
            battery_life_extension=battery_life_extension,
            coord_realignement=False # for now, should be different if we update an existing PAN
        )

        last_beacon = self.manager.get_layer('mac').database.get("macLastBeacon")
        channel = self.manager.get_layer('phy').get_channel()
        channel_page = self.manager.get_layer('phy').get_channel_page()
        beacon_payload.pan_descriptor = Dot15d4PANNetwork(last_beacon, channel_page, channel)
        self.database.set(
            "nwkOwnNetwork",
            ZigbeeNetwork(
                beacon_payload
            )
        )
        return start

    @Dot15d4Service.request("NLME-SYNC")
    def sync(self,track=False):
        """
        Implements the NLME-SYNC request.

        Allows to synchronize on the coordinator or router of a network.
        """
        macRxOnWhenIdle = self.manager.get_layer('mac').database.get("macRxOnWhenIdle")
        if not track:
            # If non enabled beacon network ...
            if macRxOnWhenIdle:
                # Get parent address (or terminate if not found)
                table = self.database.get("nwkNeighborTable")
                parent = table.get_parent()
                if parent is None:
                    return False

                # Perform a poll operation
                return self.manager.get_layer('mac').get_service('management').poll(
                    coordinator_pan_id=parent.pan_id,
                    coordinator_address=parent.address
                )
            else:
                # Enable auto request
                self.manager.get_layer('mac').database.set("macAutoRequest", True)
                # Perform a synchronization
                return self.manager.get_layer('mac').get_service('management').sync(
                    channel = self.manager.get_layer('phy').get_channel(),
                    channel_page = self.manager.get_layer('phy').get_channel_page(),
                    track_beacon=False
                )
        else:
            if macRxOnWhenIdle:
                return False
            else:
                # Enable auto request
                self.manager.get_layer('mac').database.set("macAutoRequest", True)
                # Perform a synchronization
                return self.manager.get_layer('mac').get_service('management').sync(
                    channel = self.manager.get_layer('phy').get_channel(),
                    channel_page = self.manager.get_layer('phy').get_channel_page(),
                    track_beacon=True
                )

    def on_beacon_npdu(self, pan_descriptor, beacon_payload):
        """
        Callback processing Beacon NPDU forwarded from NWK manager.
        """
        beacon_payload.pan_descriptor = pan_descriptor
        self.add_packet_to_queue(beacon_payload)


    def on_command_npdu(self, npdu, link_quality):
        """
        Callback processing Commands NPDU forwarded from NWK manager.
        """
        # Check if security header is present
        security_use = ZigbeeSecurityHeader in npdu and npdu[ZigbeeSecurityHeader].underlayer.__class__ is ZigbeeNWK
        if security_use:
            nsdu = ZigbeeNWKCommandPayload(npdu.data)
        else:
            nsdu = npdu[ZigbeeNWKCommandPayload]

        if nsdu.cmd_identifier == 6:
            self.on_rejoin_command(nsdu, npdu.source, npdu.ext_src)
        # Add packet to FIFO
        self.add_packet_to_queue(nsdu)

    def on_rejoin_command(self, nsdu, source_address, extended_source_address):
        """
        Callback processing Rejoin Commands NSDU.
        """

        table = self.database.get("nwkNeighborTable")
        candidate = table.select_by_extended_address(extended_source_address)

        capability_information = (
            int(nsdu.alternate_pan_coordinator) |
            (int(nsdu.device_type) << 1) |
            (int(nsdu.power_source) << 2) |
            (int(nsdu.receiver_on_when_idle) << 3) |
            (0 << 4) |
            (int(nsdu.security_capability) << 5) |
            (int(nsdu.allocate_address) << 7)
        )

        if candidate is None:
            network_address = source_address
        else:
            network_address = candidate.address

        permit_join = self.manager.get_layer("mac").database.get("macAssociationPermit")

        sequence_number = self.database.get("nwkSequenceNumber")
        self.database.set("nwkSequenceNumber", sequence_number+1)
        radius = self.database.get("nwkMaxDepth") * 2

        if permit_join:
            routers = table.select_routers_by_pan_id(self.database.get("nwkPANId"))

            table.update(
                network_address,
                device_type=(
                                ZigbeeDeviceType.ROUTER if
                                int(nsdu.device_type) == 1 else
                                ZigbeeDeviceType.END_DEVICE
                ),
                transmit_failure=0,
                lqi=0,
                outgoing_cost=0,
                age=0,
                extended_pan_id=self.database.get("nwkExtendedPANID"),
                logical_channel=self.manager.get_layer("phy").get_channel(),
                depth=len(routers) + 1,
                potential_parent=False,
                update_id=self.database.get("nwkUpdateId"),
                pan_id=self.database.get("nwkPANId"),
                extended_address=extended_source_address,
                rx_on_when_idle=bool((capability_information >> 3) & 1),
            )

            nwkAddressMap = self.database.get("nwkAddressMap")
            nwkAddressMap[extended_source_address] = network_address
            self.database.set("nwkAddressMap", nwkAddressMap)


            # Build rejoin response
            npdu = ZigbeeNWK(
                frametype=1,
                discover_route=0,
                seqnum=sequence_number,
                radius=radius,
                flags=["extended_src"], #, "extended_dst"],
                destination=network_address,
                source=self.database.get("nwkNetworkAddress"),
                #ext_dst=selected_parent.extended_address, # ?
                ext_src=self.database.get("nwkIeeeAddress")
            )
            nsdu = ZigbeeNWKCommandPayload(
                cmd_identifier=7,
                network_address=network_address,
                rejoin_status=0
            )


            if self.database.get("nwkSecurityLevel") != 0:
                npdu.flags.security = True
                selected_key_material = None

                security_material_set = self.database.get("nwkSecurityMaterialSet")
                for key_material in security_material_set:
                    if key_material.key_sequence_number == self.database.get("nwkActiveKeySeqNumber"):
                        selected_key_material = key_material
                        break

                if selected_key_material is None:
                    return False

                msdu = npdu / ZigbeeSecurityHeader(
                    nwk_seclevel = self.database.get("nwkSecurityLevel"),
                    source = self.database.get("nwkIeeeAddress"),
                    key_type=1,
                    key_seqnum=self.database.get("nwkActiveKeySeqNumber"),
                    fc=selected_key_material.outgoing_frame_counter,
                    data=bytes(nsdu)
                )
                crypto_manager = NetworkLayerCryptoManager(selected_key_material.key)
                msdu = crypto_manager.encrypt(msdu)
                selected_key_material.outgoing_frame_counter += 1
            else:
                msdu = npdu / nsdu

            self.manager.get_layer('mac').get_service("data").data(
                msdu,
                destination_address_mode=MACAddressMode.SHORT,
                source_address_mode=MACAddressMode.SHORT,
                destination_pan_id=self.database.get("nwkPANId"),
                destination_address=network_address,
                wait_for_ack=True
            )

            self.indicate_join(
                network_address,
                extended_source_address,
                capability_information,
                rejoin=True,
                secure_rejoin=False
            )
        else:


            # Build rejoin request
            npdu = ZigbeeNWK(
                frametype=1,
                discover_route=0,
                seqnum=sequence_number,
                radius=radius,
                flags=["extended_src"], #, "extended_dst"],
                destination=network_address,
                source=self.database.get("nwkNetworkAddress"),
                #ext_dst=selected_parent.extended_address, # ?
                ext_src=self.database.get("nwkIeeeAddress")
            )
            nsdu = ZigbeeNWKCommandPayload(
                cmd_identifier=7,
                network_address=network_address,
                rejoin_status=1
            )


            if self.database.get("nwkSecurityLevel") != 0:
                npdu.flags.security = True
                selected_key_material = None

                security_material_set = self.database.get("nwkSecurityMaterialSet")
                for key_material in security_material_set:
                    if key_material.key_sequence_number == self.database.get("nwkActiveKeySeqNumber"):
                        selected_key_material = key_material
                        break

                if selected_key_material is None:
                    return False

                msdu = npdu / ZigbeeSecurityHeader(
                    nwk_seclevel = self.database.get("nwkSecurityLevel"),
                    source = self.database.get("nwkIeeeAddress"),
                    key_type=1,
                    key_seqnum=self.database.get("nwkActiveKeySeqNumber"),
                    fc=selected_key_material.outgoing_frame_counter,
                    data=bytes(nsdu)
                )
                crypto_manager = NetworkLayerCryptoManager(selected_key_material.key)
                msdu = crypto_manager.encrypt(msdu)
                selected_key_material.outgoing_frame_counter += 1
            else:
                msdu = npdu / nsdu

            self.manager.get_layer('mac').get_service("data").data(
                msdu,
                destination_address_mode=MACAddressMode.SHORT,
                source_address_mode=MACAddressMode.SHORT,
                destination_pan_id=self.database.get("nwkPANId"),
                destination_address=network_address,
                wait_for_ack=True
            )

class NWKInterpanService(NWKService):
    """
    NWK pseudo-service forwarding InterPAN packets.
    """
    def __init__(self, manager):
        super().__init__(manager, name="nwk_interpan")

    @Dot15d4Service.request("INTRP-DATA")
    def interpan_data(
                        self,
                        asdu,
                        asdu_handle=0,
                        source_address_mode=MACAddressMode.EXTENDED,
                        destination_address_mode=MACAddressMode.SHORT,
                        destination_pan_id=0xFFFF,
                        destination_address=0xFFFF,
                        profile_id=0,
                        cluster_id=0,
                        acknowledged_transmission=False
    ):
        """
        Implements INTRP-DATA Request.

        Transmits InterPAN PDU.
        """
        # Infer delivery mode from destination address
        if destination_address == 0xFFFF:
            delivery_mode = 2
        else:
            delivery_mode = 0

        # Build scapy packet for InterPAN PDU
        data = ZigbeeNWKStub()/ZigbeeAppDataPayloadStub(
            cluster=cluster_id,
            profile=profile_id,
            delivery_mode=delivery_mode,
            data=asdu
        )

        self.manager.get_layer('mac').get_service("data").data(
            data,
            source_address_mode=source_address_mode,
            destination_pan_id=destination_pan_id,
            destination_address=destination_address,
            destination_address_mode=destination_address_mode,
            wait_for_ack=acknowledged_transmission
        )

    def on_interpan_npdu(self, pdu, destination_pan_id, destination_address, source_pan_id, source_address, link_quality):
        """
        Callback processing InterPAN NPDU forwarded by NWK manager.
        """

        # Populate profile ID & Cluster
        profile_id = pdu.profile
        cluster_id = pdu.cluster

        # Encapsulate PDU according to the selected profile (needed for correct scapy encapsulation)
        if NewZigbeeAppDataPayloadStub in pdu:
            asdu = bytes(pdu[NewZigbeeAppDataPayloadStub].payload)
        elif ZigbeeAppDataPayloadStub in pdu:
            asdu = pdu[ZigbeeAppDataPayloadStub].data

        if profile_id == 0xc05e and cluster_id == 0x1000:
            asdu = ZigbeeZLLCommissioningCluster(asdu)

        # Let's trigger an InterPAN indication
        self.indicate_interpan_data(
                                    asdu,
                                    profile_id=profile_id,
                                    cluster_id=cluster_id,
                                    destination_pan_id=destination_pan_id,
                                    destination_address=destination_address,
                                    source_pan_id=source_pan_id,
                                    source_address=source_address,
                                    link_quality=link_quality
        )

    @Dot15d4Service.indication("INTRP-DATA")
    def indicate_interpan_data(
                                self,
                                asdu,
                                profile_id=0,
                                cluster_id=0,
                                destination_pan_id=0xFFFF,
                                destination_address=0xFFFF,
                                source_pan_id=0xFFFF,
                                source_address=0xFFFF,
                                link_quality=255
    ):
        """
        Implements INTRP-DATA indication.
        """
        return (asdu,
            {
            "profile_id":profile_id,
            "cluster_id":cluster_id,
            "destination_pan_id":destination_pan_id,
            "destination_address":destination_address,
            "source_pan_id":source_pan_id,
            "source_address":source_address,
            "link_quality":link_quality
        })


@state(NWKIB)
@alias('nwk')
class NWKManager(Dot15d4Manager):
    """
    This class implements the Zigbee Network manager (NWK) and the Inter-PAN APS.
    It handles network-level operations, such as discovery, association or network initiation.

    It exposes two services providing the appropriate API.
    """
    def init(self):
        self.add_service("data", NWKDataService(self))
        self.add_service("management", NWKManagementService(self))
        self.add_service("interpan", NWKInterpanService(self))
        self.__pending_join_indication = None

    def add_key(self, key, key_sequence_number=None, outgoing_frame_counter=0):
        """
        This method adds an encryption key for Network-level Security.
        """
        networkSecurityMaterialSet = self.database.get("nwkSecurityMaterialSet")
        securityMaterial = NetworkSecurityMaterial(
            key,
            key_sequence_number=key_sequence_number,
            outgoing_frame_counter=outgoing_frame_counter
        )
        if securityMaterial not in networkSecurityMaterialSet:
            networkSecurityMaterialSet.append(securityMaterial)
        logger.info("[nwk] new security material added: %s", str(securityMaterial))
        self.database.set("nwkSecurityMaterialSet", networkSecurityMaterialSet)


    def decrypt(self, pdu):
        """
        This method decrypts a PDU if the corresponding security material is found in the database.
        """
        # Check if Security Header is present in PDU
        if ZigbeeSecurityHeader not in pdu:
            logger.info("[nwk] decryption failure - missing security header.")
            return pdu, False

        # Check the current security level
        securityLevel = self.database.get("nwkSecurityLevel")

        if securityLevel == 0:
            logger.info("[nwk] decryption failure - attempt to decrypt a pdu with nwkSecurityLevel set to 0.")
            return pdu, False

        # Check if mandatory fields are present in PDU
        if (
                not hasattr(pdu[ZigbeeSecurityHeader], "fc") or
                not hasattr(pdu[ZigbeeSecurityHeader], "source") or
                not hasattr(pdu[ZigbeeSecurityHeader], "key_seqnum")
        ):
            logger.info("[nwk] decryption failure - missing mandatory data in security header.")
            return pdu, False

        received_frame_count = pdu[ZigbeeSecurityHeader].fc
        key_sequence_number = pdu[ZigbeeSecurityHeader].key_seqnum
        sender_address = pdu[ZigbeeSecurityHeader].source

        # Select the  key from the database
        security_material_set = self.database.get("nwkSecurityMaterialSet")

        selected_key_material = None
        for key_material in security_material_set:
            if key_material.key_sequence_number == key_sequence_number:
                selected_key_material = key_material
                break
        if selected_key_material is None:
            logger.info("[nwk] decryption failure - no matching key found.")
            return pdu, False

        # Check frame counter
        if (
            sender_address in selected_key_material.incoming_frame_counters and
            received_frame_count < selected_key_material.incoming_frame_counters[sender_address] and
            self.database.get("nwkAllFresh")
        ):
            logger.info("[nwk] decryption failure - bad frame counter.")
            return pdu, False

        # Instantiate the Crypto Manager
        network_crypto_manager = NetworkLayerCryptoManager(selected_key_material.key)
        # Try to decrypt PDU
        cleartext, status = network_crypto_manager.decrypt(pdu)
        if status:
            # Increase incoming frame counter
            selected_key_material.add_incoming_frame_counter(sender_address, received_frame_count+1)
            return cleartext, True
        else:
            logger.warning("Undecoded:", bytes(pdu).hex())
            logger.info("[nwk] decryption failure - MIC not matching.")
            return pdu, False

    @source('mac', 'MCPS-DATA')
    def on_mcps_data(self, pdu, destination_pan_id, destination_address, source_pan_id, source_address, link_quality):
        """
        Callback processing MAC MCPS Data indication.
        """
        # If we have an InterPAN PDU, forwards it to the NWK InterPAN service
        if ZigbeeNWKStub in pdu and (ZigbeeAppDataPayloadStub in pdu or NewZigbeeAppDataPayloadStub in pdu):
            self.get_service("interpan").on_interpan_npdu(
                pdu,
                destination_pan_id,
                destination_address,
                source_pan_id,
                source_address,
                link_quality
            )
        # If we have a regular NWK PDU:
        elif ZigbeeNWK in pdu:
            # If PDU unencrypted, check if we allow it in current configuration
            if ZigbeeSecurityHeader not in pdu and self.database.get("nwkSecureAllFrames"):
                logger.info("[nwk] nwkSecureAllFrames attribute indicates that we only accept security enabled frames.")
                return

            # If PDU encrypted, let's try to decrypt it
            if ZigbeeSecurityHeader in pdu and pdu[ZigbeeSecurityHeader].underlayer.__class__ is ZigbeeNWK:
                decrypted, success = self.decrypt(pdu)
                if success:
                    pdu = decrypted
                else:
                    return


            # Once decrypted, forward to the right service depending on the frame type
            if pdu.frametype == 0:
                # Data PDU, forward to NWK Data Service
                self.get_service("data").on_data_npdu(pdu, link_quality)
            elif pdu.frametype == 1:
                # Management PDU, forward to NWK Management Service
                self.get_service("management").on_command_npdu(pdu, link_quality)
            else:
                # InterPAN PDU, forward to NWK InterPAN Service
                self.get_service("interpan").on_interpan_npdu(
                    pdu,
                    destination_pan_id,
                    destination_address,
                    source_pan_id,
                    source_address,
                    link_quality
                )

    @source('mac', 'MLME-ASSOCIATE')
    def on_mlme_associate(
                            self,
                            source_address,
                            capability_information,
                            security_level,
                            key_id_mode,
                            key_source,
                            key_index,
                            channel_offset,
                            hopping_sequence_id,
                            dsme_association,
                            direction,
                            allocation_order,
                            hopping_sequence_request
    ):
        """
        Callback processing MAC Association indication.
        """
        # Check if another device exists with the same address
        table = self.database.get("nwkNeighborTable")
        node = table.select_by_extended_address(source_address)
        router = (((capability_information >> 1) & 1) == 0)
        if node is not None:
            if not router:
                device_type = ZigbeeDeviceType.END_DEVICE
            else:
                device_type = ZigbeeDeviceType.ROUTER

            if device_type == node.device_type:
                success = self.get_layer('mac').get_service("management").associate_response(
                    source_address,
                    node.address,
                    association_status=MACAssociationStatus.ASSOCIATION_SUCCESSFUL
                )
                if success:
                    self.__pending_join_indication = (
                        new_address,
                        source_address,
                        capability_information,
                        False,
                        False
                    )
                return success
            else:
                table.delete(node.address, node.extended_pan_id)

        new_address = self._address_assignment(
            depth=0, # we only consider a depth of 0 (coordinator) for now
            router_node=router
        )

        if new_address is None:
            self.get_layer('mac').get_service("management").associate_response(
                source_address,
                0,
                association_status=MACAssociationStatus.PAN_AT_CAPACITY
            )
        else:
            routers = table.select_routers_by_pan_id(self.database.get("nwkPANId"))

            table.update(
                new_address,
                device_type=(
                                ZigbeeDeviceType.ROUTER if
                                router else
                                ZigbeeDeviceType.END_DEVICE
                ),
                transmit_failure=0,
                lqi=0,
                outgoing_cost=0,
                age=0,
                extended_pan_id=self.database.get("nwkExtendedPANID"),
                logical_channel=self.get_layer("phy").get_channel(),
                depth=len(routers) + 1,
                potential_parent=False,
                update_id=self.database.get("nwkUpdateId"),
                pan_id=self.database.get("nwkPANId"),
                extended_address=source_address,
                rx_on_when_idle=bool((capability_information >> 3) & 1),
            )

            nwkAddressMap = self.database.get("nwkAddressMap")
            nwkAddressMap[source_address] = new_address
            self.database.set("nwkAddressMap", nwkAddressMap)

            success = self.get_layer('mac').get_service("management").associate_response(
                source_address,
                new_address,
                association_status=MACAssociationStatus.ASSOCIATION_SUCCESSFUL
            )

            if success:
                self.__pending_join_indication = (
                    new_address,
                    source_address,
                    capability_information,
                    False,
                    False
                )

            return success

    def _address_assignment(self, depth=0, router_node=False):

        table = self.database.get("nwkNeighborTable")
        pan_id = self.database.get("nwkPANId")

        routers = table.select_routers_by_pan_id(pan_id)
        end_devices = table.select_end_devices_by_pan_id(pan_id)
        assigned_addresses = [r.address for r in routers] + [e.address for e in end_devices]

        parent = table.get_parent()
        if parent is None:
            parent_address = self.database.get("nwkNetworkAddress")
        else:
            parent_address = parent.address

        if self.database.get("nwkAddrAlloc") == 2:
            # Stochastic address assignment
            new_address = randint(1, 0xFFFD)
            while new_address in assigned_addresses:
                new_address = randint(1, 0xFFFD)
            return new_address

        elif self.database.get("nwkAddrAlloc") == 0:
            # Distributed address assignment
            cm = self.database.get("nwkMaxChildren")
            lm = self.database.get("nwkMaxDepth")
            rm = self.database.get("nwkMaxRouters")
            cskip = (
                (1 + cm * (lm - depth - 1))
                if rm == 1 else
                (1 + cm - rm  - cm * (rm ** (lm - depth - 1)))
            )

            if cskip == 0:
                return None
            else:
                if router_node:
                    new_address = parent_address + 1 + len(routers) * cskip
                else:
                    new_address = parent_address + cskip * rm + (len(end_devices)) # + 1 or not ?
            return new_address

    @source('mac', 'MLME−DATA-REQ')
    def on_mlme_data_req(self, pdu):
        if self.__pending_join_indication is not None:
            self.get_service("management").indicate_join(
                *self.__pending_join_indication
            )
            self.__pending_join_indication = None

    @source('mac', 'MLME-BEACON-NOTIFY')
    def on_mlme_beacon_notify(self, beacon_payload, pan_descriptor):
        """
        Callback processing MAC Beacon Notify indication.
        """
        # If we got bytes here, encapsulate into scapy ZigBeeBeacon
        beacon_payload = bytes(beacon_payload)

        if isinstance(beacon_payload, bytes):
            beacon_payload = ZigBeeBeacon(beacon_payload)
            # Check if this is a Zigbee beacon

        if hasattr(beacon_payload, "proto_id") and beacon_payload.proto_id in (0, 75, 92): #TODO: proto id filter at 0 does not match esp32c6 implem ? check
            # Forward it to NWK management service
            self.get_service("management").on_beacon_npdu(pan_descriptor, beacon_payload)

            # Update the neighbor table
            table = self.database.get("nwkNeighborTable")

            table.update(
                pan_descriptor.coord_addr,
                device_type=(
                                ZigbeeDeviceType.COORDINATOR if
                                pan_descriptor.coord_addr == 0 else
                                ZigbeeDeviceType.ROUTER
                ),
                transmit_failure=0,
                lqi=pan_descriptor.link_quality,
                outgoing_cost=0,
                age=0,
                extended_pan_id=beacon_payload.extended_pan_id,
                logical_channel=pan_descriptor.channel,
                depth=beacon_payload.device_depth,
                beacon_order=pan_descriptor.beacon_order,
                permit_joining=pan_descriptor.assoc_permit,
                potential_parent=True,
                update_id=beacon_payload.update_id,
                pan_id=pan_descriptor.coord_pan_id
            )

NWKManager.add(APSManager)



from whad.dot15d4.stack.database import Dot15d4Database
from whad.zigbee.stack.aps.security import APSKeyPairSet

class APSIB(Dot15d4Database):
    """
    ZigBee APSIB Database of attributes.
    """
    def reset(self):
        self.apsDeviceKeyPairSet = APSKeyPairSet(
            preinstalled_keys=[
                bytes.fromhex("814286865dc1c8b2c8cbc52e5d65d1b8"),
                b"ZigBeeAlliance09"
            ]
        )
        self.apsTrustCenterAddress = None
        self.apsSecurityTimeOutPeriod = None
        self.trustCenterPolicies = None

        self.apsDesignatedCoordinator = False
        self.apsChannelMask = 0x7fff800
        self.apsUseExtendedPANID = 0x0000000000000000
        self.apsUseInsecureJoin = False

        self.apsNonmemberRadius = 7
        self.apsCounter = 0
        self.apsUseChannel = None



"""
This module provides a scanner connector :class:`whad.esb.connector.scanner.Scanner`
for WHAD ESB devices, allowing to discover available Enhanced ShockBurst devices.

It relies on the use of the WHAD ESB protocol. It puts the hardware adapter in
pseudo-promiscuous sniffing mode and iterates over the channels to discover
Enhanced ShockBurst communications. The packets are then processed to extract
device informations.


This mode is only available with WHAD devices that support promiscuous sniffing.

This connector must be instantiated with a compatible WHAD device, as shown below:

.. code-block:: python

    device = WhadDevice.create('uart0')
    scanner = Scanner(device)

If the underlying device does not support promiscuous sniffing, this connector will raise
an :class:`UnsupportedCapability` exception.

"""
from typing import Iterator
from scapy.packet import Packet

from whad.helpers import message_filter
from whad.exceptions import UnsupportedCapability

from whad.hub.esb import PduReceived, RawPduReceived
from whad.hub.message import AbstractPacket

from .base import ESB
from ..scanning import CommunicatingDevicesDB, CommunicatingDevice

class Scanner(ESB):
    """
    ESB Observer interface for compatible WHAD device.
    """

    def __init__(self, device):
        """Instantiate scanner connector over `device`.

        :param  device: ESB WHAD device instance
        :type   device: :class:`whad.device.WhadDevice`
        """
        super().__init__(device)
        self.__db = CommunicatingDevicesDB()

        # Check device accept sniffing mode
        if not self.can_sniff():
            raise UnsupportedCapability('Sniff')

        # Stop current mode
        self.stop()

        # Switch to sniffing mode
        super().sniff(show_acknowledgements=True, address="FF:FF:FF:FF:FF")

    def start(self):
        """Start the ESB scanner.

        Calling this method resets the discovered devices database and put
        the WHAD device into ESB scanning mode.
        """
        self.__db.reset()
        super().start()

    def sniff(self, *_) -> Iterator[Packet]:
        """
        Listen and yield incoming ESB PDUs.
        """
        while True:
            if self.support_raw_pdu():
                message_type = RawPduReceived
            else:
                message_type = PduReceived

            message = self.wait_for_message(keep=message_filter(message_type))
            if issubclass(message, AbstractPacket):
                packet = message.to_packet()
                if packet is not None:
                    yield packet


    def discover_devices(self, minimal_rssi = None, filter_address = None) -> Iterator[CommunicatingDevice]:
        """
        Parse incoming packets and yield discovered devices.

        :param  minimal_rssi:       Minimal RSSI level
        :type   minimal_rssi:       float, optional
        :param  filter_address:     ESB address of a device to discover
        :type   filter_address:     :class:`whad.esb.bdaddr.ESBAddress`, optional
        """
        for packet in self.sniff():
            if minimal_rssi is None or packet.metadata.rssi > minimal_rssi:
                device = self.__db.on_device_found(
                    packet.metadata.rssi,
                    packet,
                    filter_addr=filter_address
                )
                if device is not None:
                    yield device



"""GATT Server and Client implementation
"""
import logging
from typing import Optional, List

from time import time
from queue import Queue, Empty
from struct import unpack, pack
from threading import Lock

from scapy.layers.bluetooth import ATT_Handle

from whad.ble.exceptions import HookReturnValue, HookReturnAuthentRequired,\
    HookReturnAuthorRequired, HookReturnAccessDenied, HookReturnGattError, \
    HookReturnNotFound, ConnectionLostException
from whad.ble.stack.att.constants import BleAttOpcode, BleAttErrorCode, ReadAccess, \
    WriteAccess, Authentication, Authorization, Encryption
from whad.ble.stack.att.exceptions import InsufficientEncryptionError, InsufficientAuthenticationError, InsufficientAuthorizationError, error_response_to_exc, AttErrorCode, AttError
from whad.ble.stack.gatt.message import *
from whad.ble.stack.gatt.exceptions import GattTimeoutException
from whad.ble.profile import GenericProfile
from whad.ble.stack.smp import Pairing
from whad.ble.profile.characteristic import Characteristic, Descriptor, ClientCharacteristicConfig, CharacteristicValue
from whad.ble.profile.service import Service, PrimaryService, SecondaryService, IncludeService
from whad.exceptions import WhadDeviceDisconnected
from whad.common.stack import Layer, source, alias

logger = logging.getLogger(__name__)

SUPPORTED_GROUP_TYPES = (
    UUID(0x2800), # Primary service
    UUID(0x2801), # Secondary service
    UUID(0x2802), # Include service
    UUID(0x2803), # Characteristic
    UUID(0x2900), # Characteristic extended properties
    UUID(0x2901), # Characteristic User Description
    UUID(0x2902), # Characteristic Client Configuration
    UUID(0x2903), # Server Characteristic Configuration
    UUID(0x2904), # Characteristic Presentation Format
    UUID(0x2905), # Characteristic Aggregate Format
)

def txlock(f):
    def _wrapper(self, *args, **kwargs):
        self.lock_tx()
        result = f(self, *args, **kwargs)
        self.unlock_tx()
        return result
    return _wrapper

def proclock(f):
    def _wrapper(self, *args, **kwargs):
        # Acquire GATT procedure lock
        self.procedure_start()

        # Call method and unlock GATT procedure lock
        # if an exception is raised.
        try:
            result = f(self, *args, **kwargs)
        except AttError as err:
            self.procedure_stop()
            raise err
        except GattTimeoutException as err:
            self.procedure_stop()
            raise err

        # Release GATT procedure lock
        self.procedure_stop()
        return result
    return _wrapper

@alias('gatt')
class GattLayer(Layer):
    '''Gatt client/server base class
    '''

    def configure(self, options):
        '''Configure the GATT layer
        '''
        self.__queue = Queue()
        self.__proc_lock = Lock()
        self.__tx_lock = Lock()
        self.state.terminated = False

        # Dispatch rules
        self.__handlers = {
            GattErrorResponse: self.on_error_response,

            GattFindInfoRequest: self.on_find_info_request,
            GattFindInfoResponse: self.on_find_info_response,

            GattFindByTypeValueRequest: self.on_find_by_type_value_request,
            GattFindByTypeValueResponse: self.on_find_by_type_value_response,

            GattReadByTypeRequest: self.on_read_by_type_request,
            GattReadByTypeResponse: self.on_read_by_type_response,

            GattReadByGroupTypeRequest: self.on_read_by_group_type_request,
            GattReadByGroupTypeResponse: self.on_read_by_group_type_response,

            GattReadRequest: self.on_read_request,
            GattReadResponse: self.on_read_response,

            GattReadBlobRequest: self.on_read_blob_request,
            GattReadBlobResponse: self.on_read_blob_response,

            GattReadMultipleRequest: self.on_read_multiple_request,
            GattReadMultipleResponse: self.on_read_multiple_response,

            GattWriteRequest: self.on_write_request,
            GattWriteCommand: self.on_write_command,
            GattWriteResponse: self.on_write_response,

            GattHandleValueNotification: self.on_handle_value_notification,
            GattHandleValueIndication: self.on_handle_value_indication,

            GattPrepareWriteRequest: self.on_prepare_write_request,
            GattPrepareWriteResponse: self.on_prepare_write_response,

            GattExecuteWriteRequest: self.on_execute_write_request,
            GattExecuteWriteResponse: self.on_execute_write_response,

            GattExchangeMtuRequest: self.on_exch_mtu_request,
            GattExchangeMtuResponse: self.on_exch_mtu_response,

        }

    @property
    def att(self):
        return self.get_layer('att')

    @property
    def smp(self):
        return self.get_layer('smp')

    @source('att')
    def on_att_packet(self, packet):
        '''Process ATT packets
        '''
        # Dispatch on the correct handler
        packet_type = type(packet)
        if packet_type in self.__handlers:
            self.__handlers[packet_type](packet)
        else:
            logger.error("no GATT handler for packet type %s", packet_type.__name__)

    def procedure_start(self):
        """
        Lock GATT client or server to avoid a new procedure to be initiated while
        another is actually in progress.
        """
        logger.debug('Start of GATT procedure ...')
        self.__proc_lock.acquire()

    def procedure_stop(self):
        """Unlock GATT client or server to allow other pocedures to be executed.
        """
        logger.debug('End of GATT procedure')
        self.__proc_lock.release()

    def lock_tx(self):
        """Lock GATT client or server to process a single PDU and avoid other PDUs
        to be sent in-between.
        """
        logger.debug('Start handling PDU (tx/rx)')
        self.__tx_lock.acquire()

    def unlock_tx(self):
        """Unlock GATT client or server in order to allow other PDUs to be processed.
        """
        logger.debug('End handling PDU (tx/rx)')
        self.__tx_lock.release()

    def indicate(self, characteristic):
        """Send an indication to a GATT client. Not implemented by default.
        """
        pass

    def notify(self, characteristic):
        """Send a notfication to a GATT client. Not implemented by default.
        """
        pass

    def on_gatt_message(self, message):
        """Add a GATT message into our message queue

        :param message: GATT message to add to our queue
        """
        self.__queue.put(message, block=True, timeout=None)

    def wait_for_message(self, message_clazz, timeout=10.0):
        """Wait for a specific message type or error, other messages are dropped

        :param type message_clazz: Expected message class
        :param float timeout: Timeout value (default: 30 seconds)
        """
        start_time = time()
        while (time() - start_time) < timeout:
            # Check if connection has been terminated.
            if self.state.terminated:
                logger.debug('Connection lost')
                raise ConnectionLostException(None)
            try:
                msg = self.__queue.get(block=True,timeout=timeout/10.0)
                if isinstance(msg, message_clazz) or isinstance(msg, GattErrorResponse):
                    return msg
            except Empty:
                pass
        raise GattTimeoutException


    def error(self, request, handle, reason):
        """Send ATT Error Response
        """
        self.att.error_response(
            request,
            handle,
            reason
        )

    def on_error_response(self, error: GattErrorResponse):
        self.on_gatt_message(error)

    def on_mtu_changed(self, mtu: int):
        """MTU has been updated
        """
        pass

    @txlock
    def on_find_info_request(self, packet: GattFindInfoRequest):
        '''ATT Find Information Request callback

        By default, this method generates an ATT Error Response with ATTRIBUTE_NOT_FOUND error.

        :param int start: Start handle value
        :param int end: End handle value
        '''
        logger.debug('[gatt] find_info_request, start:%d, end:%d' %(
            packet.start, packet.end
        ))
        self.error(
            BleAttOpcode.FIND_INFO_REQUEST, packet.start, BleAttErrorCode.ATTRIBUTE_NOT_FOUND
        )

    def on_find_info_response(self, packet: GattFindInfoResponse):
        '''ATT Find Information Response callback

        :param format: Information data format
        :param handles: List of handles
        '''
        pass

    @txlock
    def on_find_by_type_value_request(self, request: GattFindByTypeValueRequest):
        """ATT Find By Type Value Request callback

        :param GattFindByTypeValueRequest request: Request
        """
        logger.debug("[gatt] FindByTypeValueRequest, start: %d, type: %s", request.start, request.type_uuid)
        self.error(
            BleAttOpcode.FIND_BY_TYPE_VALUE_REQUEST, request.start, BleAttErrorCode.ATTRIBUTE_NOT_FOUND
        )

    def on_find_by_type_value_response(self, response: GattFindByTypeValueResponse):
        """ATT Find By Type Value Response callback

        :param GattFindByTypeValueResponse response: Response message
        """
        pass

    @txlock
    def on_read_by_type_request(self, request: GattReadByTypeRequest):
        """ATT Read By Type Request callback

        :param int start: Start handle value
        :param int end: End handle value
        :param uuid: Type UUID
        """
        self.error(
            BleAttOpcode.READ_BY_TYPE_REQUEST, request.start, BleAttErrorCode.ATTRIBUTE_NOT_FOUND
        )

    def on_read_by_type_response(self, response):
        """ATT Read By Type Response callback

        :param GattReadByTypeResponse response: Response
        """
        pass

    @txlock
    def on_read_by_group_type_request(self, request: GattReadByGroupTypeRequest):
        """ATT Read By Group Type Request callback

        :param int start: Start handle value
        :param int end: End handle value
        :param uuid: Type UUID
        """
        self.error(
            BleAttOpcode.READ_BY_GROUP_TYPE_REQUEST, request.start, BleAttErrorCode.ATTRIBUTE_NOT_FOUND
        )


    def on_read_by_group_type_response(self, response):
        """ATT Read By Group Type Response callback

        :param int item_length: Item length
        :param data: List of items
        """
        pass

    @txlock
    def on_read_request(self, request: GattReadRequest):
        """ATT Read Request callback

        :param int handle: Attribute handle
        """
        self.error(
            BleAttOpcode.READ_REQUEST, request.handle, BleAttErrorCode.INVALID_HANDLE
        )

    def on_read_response(self, response: GattReadResponse):
        """ATT Read Response callback

        :param value: Attribute value
        """
        pass

    @txlock
    def on_read_blob_request(self, request: GattReadBlobRequest):
        """ATT Read Blob Request callback

        :param int handle: Attribute handle
        :param int offset: Offset of the first byte of data to read
        """
        self.error(
            BleAttOpcode.READ_BLOB_REQUEST, request.handle, BleAttErrorCode.INVALID_HANDLE
        )

    def on_read_blob_response(self, response: GattReadBlobResponse):
        """ATT Read Blob Response callback

        :param value: Attribute value
        """
        pass

    @txlock
    def on_read_multiple_request(self, request: GattReadMultipleRequest):
        """ATT Read Multiple Request callback

        :param handles: List of handles
        """
        self.error(
            BleAttOpcode.READ_MULTIPLE_REQUEST, request.handles[0], BleAttErrorCode.INVALID_HANDLE
        )

    def on_read_multiple_response(self, response: GattReadMultipleResponse):
        """ATT Read Multiple Response callback

        :param values: Multiple Attribute values
        """
        pass


    @txlock
    def on_write_request(self, request: GattWriteRequest):
        """ATT Write Request callback

        :param int handle: Attribute handle
        :param data: Attribute value
        """
        self.error(
            BleAttOpcode.WRITE_REQUEST, request.handle, BleAttErrorCode.INVALID_HANDLE
        )

    def on_write_response(self, response):
        """ATT Write Response callback
        """
        pass

    @txlock
    def on_write_command(self, request: GattWriteCommand):
        """ATT Write Command callback

        :param int handle: Attribute handle
        :param data: Attribute data
        """
        self.error(
            BleAttOpcode.WRITE_COMMAND, request.handle, BleAttErrorCode.INVALID_HANDLE
        )

    def on_handle_value_notification(self, notification: GattHandleValueNotification):
        """ATT Handle Value Notification

        :param int handle: Attribute handle
        :param value: Attribute value
        """
        pass

    @txlock
    def on_handle_value_indication(self, indication: GattHandleValueIndication):
        """ATT Handle Value Indication

        :param int handle: Attribute handle
        :param value: Attribute value
        """
        self.att.handle_value_confirmation()

    @txlock
    def on_prepare_write_request(self, request: GattPrepareWriteRequest):
        """ATT Prepare Write request callback

        :param int handle: Attribute handle
        :param int offset: Data offset
        :param data: Attribute data
        """
        self.error(
            BleAttOpcode.PREPARE_WRITE_REQUEST, request.handle, BleAttErrorCode.INVALID_HANDLE
        )

    def on_prepare_write_response(self, response: GattPrepareWriteResponse):
        """ATT Prepare Write Response Callback

        :param int handle: Attribute handle
        :param int offset: Data offset
        :param data: Attribute data
        """
        pass

    def on_execute_write_request(self, request: GattExecuteWriteRequest):
        """ATT Execute Write Request callback

        :param int flags: Flags
        """
        pass

    def on_execute_write_response(self, response: GattExecuteWriteResponse):
        """ATT Execute Write Response callback
        """
        pass

    def on_exch_mtu_request(self, request):
        """ATT MTU exchange request

        :param request: request from remote device
        """
        pass

    def on_exch_mtu_response(self, response):
        """ATT MTU exchange response

        :param GattExchangeMtuResponse response: response from remote device
        """
        self.on_gatt_message(response)


class GattClient(GattLayer):

    def __init__(self, parent=None, layer_name=None, options={}):
        super().__init__(parent=parent, layer_name=layer_name, options=options)
        self.__model = None
        self.__notification_callbacks = {}

    def configure(self, options):
        '''Configure GATT client.
        '''
        super().configure(options)


    @property
    def model(self):
        return self.__model

    def set_model(self, model):
        '''Set the underlying GATT profile.
        '''
        if isinstance(model, GenericProfile):
            self.__model = model

    def set_client_model(self, model):
        '''This method is used with GattClientServer to specify the GATT
        client device profile.
        '''
        if isinstance(model, GenericProfile):
            self.__model = model


    ###################################
    # Supported response handlers
    ###################################

    def on_read_by_group_type_response(self, response: GattReadByGroupTypeResponse):
        """ATT Read By Group Type Response callback
        """
        self.on_gatt_message(response)

    def on_find_info_response(self, response):
        """ATT Find Information Response callback

        :param format: Information data format
        :param handles: List of handles
        """
        self.on_gatt_message(response)

    def on_read_response(self, response):
        """ATT Read Response callback

        :param value: Attribute value
        """
        self.on_gatt_message(response)

    def on_read_by_type_response(self, response: GattReadByTypeResponse):
        """ATT Read By Type Response callback

        :param GattReadByTypeResponse response: Response
        """
        self.on_gatt_message(response)

    def on_read_blob_response(self, response: GattReadBlobResponse):
        """ATT Read Blob Response callback

        :param value: Attribute value
        """
        self.on_gatt_message(response)

    def on_write_response(self, response):
        """ATT Write Response callback
        """
        self.on_gatt_message(response)

    def on_prepare_write_response(self, response):
        """ATT Prepare Write Response Callback

        :param GattPrepareWriteResponse response: Response message
        """
        self.on_gatt_message(response)

    def on_execute_write_response(self, response):
        """ATT Execute Write Response Callback

        :param GattExecuteWriteResponse response: Response message
        """
        self.on_gatt_message(response)

    def on_find_by_type_value_response(self, response: GattFindByTypeValueResponse):
        """ATT Find By Type Value Response callback

        :param GattFindByTypeValueResponse response: Response message
        """
        self.on_gatt_message(response)

    def on_handle_value_notification(self, notification):
        """ATT Handle Value Notification

        :param int handle: Attribute handle
        :param value: Attribute value
        """
        if notification.handle in self.__notification_callbacks:
            self.__notification_callbacks[notification.handle](
                notification.handle,
                notification.value,
                indication=False
            )

    @txlock
    def on_handle_value_indication(self, notification):
        """ATT Handle Value Indication

        :param int handle: Attribute handle
        :param value: Attribute value
        """
        if notification.handle in self.__notification_callbacks:
            self.__notification_callbacks[notification.handle](
                notification.handle,
                notification.value,
                indication=True
            )
        self.att.handle_value_confirmation()

    @txlock
    def on_exch_mtu_request(self, request):
        """Received an MTU exchange request
        """
        # We do accept MTU exchange request, save server MTU and answer
        self.att.set_server_mtu(request.mtu)
        self.att.set_client_mtu(request.mtu)
        self.att.exch_mtu_response(self.att.get_client_mtu())

        # Notify our connector about this MTU update
        logging.debug("[gatt] notifying connector new server mtu: %d", request.mtu)
        self.att.notify_server_mtu()

    def on_exch_mtu_response(self, response):
        """ATT MTU exchange response

        :param GattExchangeMtuResponse response: response from remote device
        """
        self.on_gatt_message(response)


    ###################################
    # GATT procedures
    ###################################

    def register_notification_callback(self, handle, cb):
        self.__notification_callbacks[handle] = cb

    def unregister_notification_callback(self, handle):
        if handle in self.__notification_callbacks:
            del self.__notification_callbacks[handle]

    def clear_notification_callbacks(self):
        """Clear all notification callbacks.
        """
        self.__notification_callbacks = {}

    @proclock
    def discover_primary_service_by_uuid(self, uuid):
        """Discover a primary service by its UUID.

        :param UUID uuid: Service UUID
        :return: Service if service has been found, None otherwise
        """
        # Send FindByTypeValueRequest
        self.lock_tx()
        self.att.find_by_type_value_request(
            1,
            0xFFFF,
            0x2800,
            uuid.packed
        )
        self.unlock_tx()

        msg = self.wait_for_message(GattFindByTypeValueResponse)
        if isinstance(msg, GattFindByTypeValueResponse):
            for item in msg:
                return PrimaryService(
                    uuid=uuid,
                    handle=item.handle,
                    end_handle=item.end
                )
        elif isinstance(msg, GattErrorResponse):
            if msg.reason == AttErrorCode.ATTR_NOT_FOUND:
                return None
            else:
                raise error_response_to_exc(msg.reason, msg.request, msg.handle)

    @proclock
    def discover_primary_services(self, start: int = 1):
        """Discover remote Primary Services.

        This function will yield every discovered primary service.
        """
        # List primary services handles
        handle = start
        while True:
            # Send a Read By Group Type Request
            self.lock_tx()
            self.att.read_by_group_type_request(
                handle,
                0xFFFF,
                0x2800
            )
            self.unlock_tx()

            msg = self.wait_for_message(GattReadByGroupTypeResponse)
            if isinstance(msg, GattReadByGroupTypeResponse):
                for item in msg:
                    yield PrimaryService(
                        UUID(item.value),
                        handle=item.handle,
                        end_handle=item.end
                    )
                    handle = item.end

                    if handle == 0xFFFF:
                        return

                handle += 1
            elif isinstance(msg, GattErrorResponse):
                if msg.reason == AttErrorCode.ATTR_NOT_FOUND:
                    break
                else:
                    error_response_to_exc(msg.reason, msg.request, msg.handle)

    @proclock
    def discover_secondary_services(self):
        """Discover remote Secondary Services.
        """
        # List primary services handles
        handle = 1
        while True:
            # Send a Read By Group Type Request
            self.lock_tx()
            self.att.read_by_group_type_request(
                handle,
                0xFFFF,
                0x2801
            )
            self.unlock_tx()

            msg = self.wait_for_message(GattReadByGroupTypeResponse)
            if isinstance(msg, GattReadByGroupTypeResponse):
                for item in msg:
                    yield SecondaryService(
                        uuid=UUID(item.value),
                        handle=item.handle,
                        end_handle=item.end
                    )
                    handle = item.end

                    if handle == 0xFFFF:
                        return

                handle += 1
            elif isinstance(msg, GattErrorResponse):
                if msg.reason == AttErrorCode.ATTR_NOT_FOUND:
                    break
                else:
                    error_response_to_exc(msg.reason, msg.request, msg.handle)

    @proclock
    def discover_characteristics(self, service: Service, save_values: bool = False, start: Optional[int] = None):
        """
        Discover a given service's characteristics and automatically add them to the specified service object.

        Discovered characteristics are also yielded by this method, providing a way to determine what characteristics
        have been discovered.

        :param service: Service to discover
        :type  service: Service
        :param save_values: When set to `True`, will read the discovered characteristics' values and save them
        :type  save_values: bool, optional
        :param start: If set, discover characteristics with handle in range [`start`, 0xFFFF]
        :type  start: int, optional
        :return: Iterator for discovered characteristics enumeration
        :rtype: Iterator[Characteristic]
        """
        logger.debug("discover characteristics for service %s", service.uuid)
        logger.debug("discover characteristics from handle %d to %d",
            service.handle, service.end_handle
        )
        if isinstance(service, Service):
            if start is not None:
                handle = start
            else:
                handle = service.handle
        else:
            return

        while handle <= service.end_handle:
            logger.debug("service end handle is: %d", service.end_handle)
            self.lock_tx()
            self.att.read_by_type_request(
                handle,
                service.end_handle,
                0x2803
            )
            self.unlock_tx()

            msg = self.wait_for_message(GattReadByTypeResponse)
            if isinstance(msg, GattReadByTypeResponse):
                for item in msg:
                    charac_properties = item.value[0]
                    #charac_handle = unpack('<H', item.value[1:3])[0]
                    charac_handle = item.handle
                    charac_value_handle = unpack('<H', item.value[1:3])[0]
                    charac_uuid = UUID(item.value[3:])
                    charac = Characteristic(
                        handle=charac_handle,
                        uuid=charac_uuid,
                        properties=charac_properties
                    )
                    charac.value_handle = charac_value_handle
                    charac.service = service

                    # Add characteristic as soon as we discovered it and set its value
                    service.add_characteristic(charac)

                    # Read value if requested and characteristic is readable
                    if save_values and (charac_properties & 0x02) > 0:
                        try:
                            charac.value = self.read(charac_value_handle)
                        except (AttError, GattTimeoutException):
                            charac.value = None

                    handle = charac.handle+2
                    logger.debug("found characteristic %s with handle %d", charac_uuid, charac_value_handle)
                    yield charac

            elif isinstance(msg, GattErrorResponse):
                if msg.reason == AttErrorCode.ATTR_NOT_FOUND:
                    break
                else:
                    error_response_to_exc(msg.reason, msg.request, msg.handle)

    @proclock
    def discover_characteristic_descriptors(self, characteristic):
        """Find characteristic descriptor
        """
        if isinstance(characteristic, Characteristic):
            handle = characteristic.value_handle + 1
            end_handle = self.__model.find_characteristic_end_handle(characteristic.handle)
            while handle <= end_handle:
                self.lock_tx()
                self.att.find_info_request(
                    handle,
                    end_handle
                )
                self.unlock_tx()

                msg = self.wait_for_message(GattFindInfoResponse)
                if isinstance(msg, GattFindInfoResponse):
                    for descriptor in msg:
                        handle = descriptor.handle

                        # End discovery if returned handle is Ending Handle (0xFFFF)
                        if handle == 0xFFFF:
                            return

                        yield(descriptor)
                elif isinstance(msg, GattErrorResponse):
                    if msg.reason == AttErrorCode.ATTR_NOT_FOUND:
                        break
                    else:
                        error_response_to_exc(msg.reason, msg.request, msg.handle)

                handle += 1
    @proclock
    def read_characteristic_by_uuid(self, uuid: UUID, start: int = 1, end: int = 0xffff) -> Optional[List[CharacteristicValue]]:
        """Read a characteristic by its UUID.

        :param UUID uuid: Characteristic UUID
        :return: Characteristic if characteristic has been found, None otherwise
        """
        output = []

        if uuid.type == UUID.TYPE_16:
            self.lock_tx()
            self.att.read_by_type_request(start, end, uuid.packed)
            self.unlock_tx()
        elif uuid.type == UUID.TYPE_128:
            uuid1 = uuid.packed[0:4]
            uuid2 = uuid.packed[4:8]
            self.lock_tx()
            self.att.read_by_type_request_128(start, end, uuid1, uuid2)
            self.unlock_tx()

        msg = self.wait_for_message(GattReadByTypeResponse)
        if isinstance(msg, GattReadByTypeResponse):
            for item in msg:
                output.append(CharacteristicValue(uuid, item.handle, item.value))
        elif isinstance(msg, GattErrorResponse):
            if msg.reason == AttErrorCode.ATTR_NOT_FOUND:
                return None
            else:
                raise error_response_to_exc(msg.reason, msg.request, msg.handle)

        # Return found characteristic values
        return output

    def get_descriptor(self, characteristic: Characteristic, uuid: UUID, handle: int) -> Descriptor:
        """Read a characteristic descriptor identified by its handle.

        @param handle: Descriptor handle
        @type handle: int
        @return Characteristic descriptor
        @rtype CharacteristicDescriptor
        """
        # Read descriptor value
        desc_value = self.read(handle)

        # Return descriptor object based on value and UUID
        return Descriptor.from_uuid(characteristic, handle,
                                                      uuid, desc_value)

    def discover(self, save_values: bool = False):
        """Discover device services, characteristics and descriptors.

        This method performs a GATT server discovery procedure, in three distinct steps:

        1. Enumerate all primary services,
        2. For each service, enumerate all its characteristics and read values if needed
        3. For each characteristic, enumerate all its descriptors

        :param save_values: If `True`, save every characteristic's value in the generated profile
        :type  save_values: bool
        """
        # Discover services
        services = []
        for service in self.discover_primary_services():
            services.append(service)

            # Add service into model
            self.__model.add_service(service)


        # Discover characteristics for each service, skipping already discovered services
        for service in services:
            # Enumerating characteristics
            for characteristic in self.discover_characteristics(service, save_values=save_values):
                    logger.debug("found new characteristic %s with handle %s", characteristic.uuid,
                                 characteristic.handle)

            # Update service into model
            self.__model.add_service(service)


        # Searching for descriptors
        for service in services:
            for characteristic in service.characteristics():
                for descriptor in self.discover_characteristic_descriptors(characteristic):
                    try:
                        desc = self.get_descriptor(characteristic, descriptor.uuid, descriptor.handle)
                        if desc is not None:
                            # Add descriptor to characteristic
                            characteristic.add_descriptor(desc)

                            # Update model: register descriptor into the model's attribute database
                            self.__model.register_attribute(desc)
                    except (InsufficientAuthenticationError, InsufficientAuthorizationError, InsufficientEncryptionError):
                        pass

    @proclock
    def read(self, handle):
        """Read a characteristic or a descriptor.

        :param int handle: Handle of the attribute to read (descriptor or characteristic)
        """
        self.lock_tx()
        self.att.read_request(gatt_handle=handle)
        self.unlock_tx()

        msg = self.wait_for_message(GattReadResponse)
        if isinstance(msg, GattReadResponse):
            return msg.value
        elif isinstance(msg, GattErrorResponse):
            raise error_response_to_exc(msg.reason, msg.request, msg.handle)

    @proclock
    def read_blob(self, handle, offset=0):
        """Read a characteristic or a descriptor starting from `offset`.

        :param int handle: Handle of the characteristic value or descriptor to read.
        :param int offset: Start reading from this offset value (default: 0)
        """
        self.lock_tx()
        self.att.read_blob_request(handle, offset)
        self.unlock_tx()

        msg = self.wait_for_message(GattReadBlobResponse)
        if isinstance(msg, GattReadBlobResponse):
            return msg.value
        elif isinstance(msg, GattErrorResponse):
            raise error_response_to_exc(msg.reason, msg.request, msg.handle)

    @proclock
    def read_long(self, handle):
        """Read a long characteristic or descriptor

        :param int handle: Handle of the attribute to read (descriptor or characteristic)
        """
        local_mtu = self.att.get_server_mtu()

        value=b''
        offset=0
        while True:
            # Send a Read request for first part, then ReadBlob request
            # for additional parts.
            self.lock_tx()
            if value == b'':
                self.att.read_request(handle)
            else:
                self.att.read_blob_request(handle, offset)
            self.unlock_tx()

            msg = self.wait_for_message((GattReadResponse, GattReadBlobResponse))
            if isinstance(msg, (GattReadResponse, GattReadBlobResponse)):
                if len(msg.value) < (local_mtu - 1):
                    value += msg.value
                    break
                else:
                    value += msg.value
                    offset += len(msg.value)
            elif isinstance(msg, GattErrorResponse):
                raise error_response_to_exc(msg.reason, msg.request, msg.handle)
        return value

    @proclock
    def write(self, handle, value):
        """Write data to a characteristic or a descriptor

        :param int handle: Target characteristic or descriptor handle
        :param bytes value: Data to write
        """
        local_mtu = self.att.get_server_mtu()

        # Check if data to write is longer than MTU
        if len(value) > (local_mtu - 3):
            # Redirect to write_long_nolock()
            return self.write_long_nolock(handle, value)
        else:
            # If data can be sent in a single write request, just send it :)
            self.lock_tx()
            self.att.write_request(
                handle,
                value
            )
            self.unlock_tx()

            msg = self.wait_for_message(GattWriteResponse)
            if isinstance(msg, GattWriteResponse):
                return True
            elif isinstance(msg, GattErrorResponse):
                raise error_response_to_exc(msg.reason, msg.request, msg.handle)

    @proclock
    def write_command(self, handle, value):
        """Write data to a characteristic or a descriptor, do not expect an answer.

        :param int handle: Target characteristic or descriptor handle
        :param bytes value: Data to write
        """
        self.lock_tx()
        self.att.write_command(
            handle,
            value
        )
        self.unlock_tx()

        # Write command does not cause the GATT server to return a response.
        return True

    @proclock
    def write_long(self, handle, value):
        """Write a long value into a characteristic

        This function wraps a call to `write_long_nolock()` to ensure this procedure
        is not started while another GATT procedure is in progress.
        """
        return self.write_long_nolock(handle, value)

    def write_long_nolock(self, handle, value):
        """Write long data (size > ATT_MTU-2) to a characteristic value.

        This method uses Prepared Write requests as described in Vol 3, Part G,
        section 4.9.5.

        :param int handle: Target characteristic value handle
        :param bytes value: Data to write
        """
        # We need to determine how many prepared write requests we should issue
        # to the target device
        data_len = len(value)
        local_mtu = self.att.get_server_mtu()

        nb_chunks = int(data_len / (local_mtu - 5))
        if nb_chunks % (local_mtu - 5) > 0:
            nb_chunks += 1
        chunk_size = local_mtu - 5

        # Send prepared write requests
        offset = 0
        for i in range(nb_chunks):
            self.lock_tx()
            self.att.prepare_write_request(handle, offset, value[offset:offset + chunk_size])
            self.unlock_tx()

            msg = self.wait_for_message(GattPrepareWriteResponse)
            if isinstance(msg, GattPrepareWriteResponse):
                if msg.value is not None:
                    offset += len(msg.value)
                else:
                    break
            elif isinstance(msg, GattErrorResponse):
                raise error_response_to_exc(msg.reason, msg.request, msg.handle)

        # Execute write request
        self.lock_tx()
        self.att.execute_write_request(1)
        self.unlock_tx()

        msg = self.wait_for_message(GattExecuteWriteResponse)
        if isinstance(msg, GattExecuteWriteResponse):
            return True
        elif isinstance(msg, GattErrorResponse):
            raise error_response_to_exc(msg.reason, msg.request, msg.handle)

    @proclock
    def find_characteristics_by_uuid(self, uuid, start=1, end=0xFFFF) -> List[Characteristic]:
        """Find a characteristic given its UUID if its handle is comprised in a given range.

        :param UUID uuid: Characteristic UUID
        :param int start: Start handle value
        :param int end: End handle value
        """
        output = []
        handle = start
        while handle <= end:
            self.lock_tx()
            self.att.read_by_type_request(
                handle,
                end,
                0x2803
            )
            self.unlock_tx()

            msg = self.wait_for_message(GattReadByTypeResponse)
            if isinstance(msg, GattReadByTypeResponse):
                for item in msg:
                    # Retrieve characteristic handle and UUID
                    charac_handle = item.handle
                    charac_uuid = UUID(item.value[3:])

                    # If characteristic has been found, add it to our output
                    if charac_uuid == uuid:
                        # Parse remaining fields
                        charac_properties = item.value[0]
                        charac_value_handle = unpack('<H', item.value[1:3])[0]

                        # Create a new Characteristic object
                        charac = Characteristic(
                            handle=charac_handle,
                            uuid=charac_uuid,
                            properties=charac_properties
                        )
                        charac.value_handle = charac_value_handle

                        # Add to our found characteristics
                        output.append(charac)

                    handle = charac_handle+2


            elif isinstance(msg, GattErrorResponse):
                if msg.reason == AttErrorCode.ATTR_NOT_FOUND:
                    break
                else:
                    error_response_to_exc(msg.reason, msg.request, msg.handle)

        # Return found characteristic(s)
        return output

    @proclock
    def set_mtu(self, mtu):
        '''Set (G)ATT client MTU (must be >= ATT_MTU, i.e. 23).

        This method is exposed in GattClient but really interact with the underlying
        ATT layer, and checks if we receive an answer from the remote device.

        :param int mtu: ATT MTU to use
        :return int: remote device MTU
        '''
        if mtu >= 23:
            self.lock_tx()
            self.att.exch_mtu_request(mtu)
            self.unlock_tx()

            msg = self.wait_for_message(GattExchangeMtuResponse)
            if isinstance(msg, GattExchangeMtuResponse):
                # Update client and server MTU
                self.att.set_client_mtu(mtu)
                self.att.set_server_mtu(msg.mtu)

                # Notify our connector that the remote server has
                # changed its MTU
                logging.debug("[gatt client] notifying received server mtu: %d", msg.mtu)
                self.att.notify_server_mtu()
                return msg.mtu
            elif isinstance(msg, GattErrorResponse):
                raise error_response_to_exc(msg.reason, msg.request, msg.handle)
        else:
            return None

    def on_mtu_changed(self, mtu):
        """MTU has changed, notify client.
        """
        pass

    def services(self):
        return self.__model.services()

    def on_terminated(self):
        # Process termination.
        #super().on_terminated()

        # Remove all notification/indication callbacks
        self.clear_notification_callbacks()


class GattServer(GattLayer):
    """
    BLE GATT server
    """

    def __init__(self, parent=None, layer_name=None, options={}):
        super().__init__(parent=parent, layer_name=layer_name, options=options)
        self.__server_model = None

        # Prepared write queues
        self.__write_queues = {}

        # Subscribed characteristics
        self.__subscribed_characs = []

    @property
    def server_model(self):
        """Retrieve the current server model
        """
        return self.__server_model

    def set_model(self, model):
        '''Set the GATT server underlying profile.
        '''
        self.__server_model = model

    def set_server_model(self, model):
        '''This method is used with GattClientServer to specify the GATT
        server profile.
        '''
        # Note: redundant with set_server_model
        self.__server_model = model

    def configure(self, options):
        '''Configure GATT client.
        '''
        super().configure(options)


    ###################################
    # Supported response handlers
    ###################################


    ###################################
    # GATT procedures
    ###################################

    def set_mtu(self, mtu):
        '''Set (G)ATT server MTU (must be >= ATT_MTU, i.e. 23).

        This method is exposed in GattServer but really interact with the underlying
        ATT layer, and checks if we receive an answer from the remote device.

        :param int mtu: ATT MTU to use
        :return int: remote device MTU
        '''
        if mtu >= 23:
            self.lock_tx()
            self.att.exch_mtu_request(mtu)
            self.unlock_tx()

            msg = self.wait_for_message(GattExchangeMtuResponse)
            if isinstance(msg, GattExchangeMtuResponse):
                # Save client and server MTU
                self.att.set_server_mtu(mtu)
                self.att.set_client_mtu(msg.mtu)

                # Notify our connector that the remote client has changed
                # its ATT MTU
                #logging.debug("[gatt server] notifying received client mtu: %d", msg.mtu)
                #self.att.notify_client_mtu()
                return msg.mtu
            elif isinstance(msg, GattErrorResponse):
                # If an error has been received, that means the MTU exchange
                # procedure failed. We keep the same MTU.
                return None
        else:
            return None

    @proclock
    def notify(self, characteristic):
        """Sends a notification to a GATT client for a given characteristic.

        :param Characteristic characteristic: Characteristic to notify the GATT client about.
        """
        try:
            local_mtu = self.att.get_client_mtu()

            # Call model callback
            service = self.server_model.find_service_by_characteristic_handle(characteristic.handle)
            self.server_model.on_notification(
                service,
                characteristic,
                characteristic.value[:local_mtu-3]
            )

            # Send notification
            self.att.handle_value_notification(
                characteristic.value_handle,
                characteristic.value[:local_mtu-3]
            )
        except HookReturnValue as value_override:
            # Return overriden value
            self.att.handle_value_notification(
                characteristic.value_handle,
                value_override.value[:local_mtu-3]
            )

    @proclock
    def indicate(self, characteristic):
        """Sends an indication to a GATT client for a given characteristic.

        :param Characteristic characteristic: Characteristic to notify the GATT client about.
        """
        try:
            local_mtu = self.att.get_client_mtu()

            # Call model callback
            service = self.server_model.find_service_by_characteristic_handle(characteristic.handle)
            self.server_model.on_indication(
                service,
                characteristic,
                characteristic.value[:local_mtu-3]
            )

            # Send notification
            self.att.handle_value_indication(
                characteristic.value_handle,
                characteristic.value[:local_mtu-3]
            )
        except HookReturnValue as value_override:
            # Return overriden value
            self.att.handle_value_indication(
                characteristic.value_handle,
                value_override.value[:local_mtu-3]
            )

    @txlock
    def on_find_info_request(self, request):
        """Find information request
        """
        # Make sure the requested start handle is valid:
        # - start handle must not be 0
        # - start handle must be lower or equal to end handle
        if request.start == 0 or request.start > request.end:
            self.error(
                BleAttOpcode.FIND_INFO_REQUEST,
                request.start,
                BleAttErrorCode.INVALID_HANDLE
            )
        else:
            # List attributes by type UUID, sorted by handles
            attrs = {}
            attrs_handles = []
            for attribute in self.server_model.find_objects_by_range(request.start, request.end):
                attrs[attribute.handle] = attribute
                attrs_handles.append(attribute.handle)
            attrs_handles.sort()

            # If we have at least one item to return
            if len(attrs_handles) > 0:

                # Get MTU
                mtu = self.att.get_client_mtu()

                # Get item size (UUID size + 2)
                uuid_size = len(attrs[attrs_handles[0]].type_uuid.packed)
                if uuid_size == 2:
                    item_format = 1
                else:
                    item_format = 2
                item_size = uuid_size + 2
                max_nb_items = int((mtu - 2) / item_size)

                # Create our datalist
                datalist = GattAttributeDataList(item_size)

                # Iterate over items while UUID size matches and data fits in MTU
                for i in range(max_nb_items):
                    if i < len(attrs_handles):
                        handle = attrs_handles[i]
                        attr_obj = attrs[handle]
                        if len(attr_obj.type_uuid.packed) == uuid_size:
                            datalist.append(
                                GattHandleUUIDItem(
                                    attr_obj.handle,
                                    attr_obj.type_uuid
                                )
                            )
                    else:
                        break

                # Once datalist created, send answer
                datalist_raw = datalist.to_bytes()
                self.att.find_info_response(item_format, datalist_raw)
            else:
                self.error(
                   BleAttOpcode.FIND_INFO_REQUEST,
                   request.start,
                   BleAttErrorCode.ATTRIBUTE_NOT_FOUND
                )

    @txlock
    def on_find_by_type_value_request(self, request: GattFindByTypeValueRequest):
        """ATT Find By Type Value Request callback

        :param GattFindByTypeValueRequest request: Request
        """
        # Make sure the requested start handle is valid:
        # - start handle must not be 0
        # - start handle must be lower or equal to end handle
        if request.start == 0 or request.start > request.end:
            self.error(
                BleAttOpcode.FIND_BY_TYPE_VALUE_REQUEST,
                request.start,
                BleAttErrorCode.INVALID_HANDLE
            )
        else:
            # List attributes by type UUID, sorted by handles
            attrs = {}
            attrs_handles = []
            for attribute in self.server_model.find_objects_by_range(request.start, request.end):
                attrs[attribute.handle] = attribute
                attrs_handles.append(attribute.handle)
            attrs_handles.sort()

            # Loop on attributes and return the attributes with a value that matches the request value
            matching_attrs = []
            for handle in attrs_handles:
                # Retrieve attribute based on handle
                attr = attrs[handle]

                if UUID(request.type_uuid) != attr.type_uuid:
                    continue

                # If attribute is a characteristic value or a descriptor, we make sure the characteristic
                # is readable before matching its value with the request value
                if isinstance(attr, CharacteristicValue) or isinstance(attr, Descriptor):
                    # Find characteristic end handle
                    if attr.value == request.attr_data:
                        matching_attrs.append((handle, attr.characteristic.end_handle))
                else:
                    # PrimaryService and SecondaryService are grouping types
                    if isinstance(attr, PrimaryService) or isinstance(attr, SecondaryService):
                        if attr.value == request.value:
                            matching_attrs.append((handle, attr.end_handle))
                    else:
                        if attr.value == request.value:
                            matching_attrs.append((handle, handle))

            # If we have found at least one attribute that matches the request, return a
            # FindByTypeValueResponse PDU
            if len(matching_attrs) > 0:
                # Build the response
                mtu = self.att.get_client_mtu()
                max_nb_items = int((mtu - 1) / 4)

                # Create our datalist
                handles_list = []

                # Iterate over items while UUID size matches and data fits in MTU
                for i in range(max_nb_items):
                    if i < len(matching_attrs):
                        handle, end_handle = matching_attrs[i]
                        handles_list.append(
                            ATT_Handle(
                                handle=handle,
                                value=end_handle
                            )
                        )
                    else:
                        break

                # Once datalist created, send answer
                self.att.find_by_type_value_response(handles_list)
            else:
                # Attribute not found
                self.error(
                   BleAttOpcode.FIND_BY_TYPE_VALUE_REQUEST,
                   request.start,
                   BleAttErrorCode.ATTRIBUTE_NOT_FOUND
                )


    @txlock
    def on_read_request(self, request):
        """Read attribute value (if any)

        :param int handle: Characteristic or descriptor handle
        """
        if request.handle == 0:
            return self.error(
                BleAttOpcode.READ_REQUEST,
                request.handle,
                BleAttErrorCode.INVALID_HANDLE
            )

        try:
            local_mtu = self.att.get_client_mtu()

            # Search attribute by handle and send respons
            attr = self.server_model.find_object_by_handle(request.handle)

            # Ensure attribute is a readable characteristic value or a descriptor
            if isinstance(attr, CharacteristicValue):

                # Check characteristic is readable
                charac = self.server_model.find_object_by_handle(request.handle - 1)

                conn_handle = self.get_layer('l2cap').get_conn_handle()
                if charac.check_security_property(ReadAccess, Authentication):
                    print("[i] authentication required for read access !")
                    if not self.get_layer('ll').state.is_authenticated(conn_handle):
                        self.error(
                            BleAttOpcode.READ_REQUEST,
                            request.handle,
                            BleAttErrorCode.INSUFFICIENT_AUTHENT
                        )
                        return
                if charac.check_security_property(ReadAccess, Encryption):
                    print("[i] encryption required for read access !")
                    if not self.get_layer('ll').state.is_encrypted(conn_handle):
                        self.error(
                            BleAttOpcode.READ_REQUEST,
                            request.handle,
                            BleAttErrorCode.INSUFFICIENT_ENCRYPTION
                        )
                        return
                if charac.check_security_property(ReadAccess, Authorization):
                    print("[i] authorization required for read access !")
                    # TODO: not supported for now
                    self.error(
                        BleAttOpcode.READ_REQUEST,
                        request.handle,
                        BleAttErrorCode.INSUFFICIENT_AUTHOR
                    )
                    return
                if charac.readable():
                    try:
                        service = self.server_model.find_service_by_characteristic_handle(charac.handle)
                        self.server_model.on_characteristic_read(
                            service,
                            charac,
                            0,
                            local_mtu - 1
                        )

                        # Make sure the returned value matches the boundaries
                        value = charac.value[:local_mtu - 1]

                        self.att.read_response(
                            value
                        )
                    except HookReturnValue as force_value:
                        # Make sure the returned value matches the boundaries
                        value = force_value.value[:local_mtu - 1]

                        self.att.read_response(
                            value
                        )
                    except HookReturnAuthentRequired:
                        self.error(
                            BleAttOpcode.READ_REQUEST,
                            request.handle,
                            BleAttErrorCode.INSUFFICIENT_AUTHENT
                        )
                    except HookReturnAuthorRequired:
                        self.error(
                            BleAttOpcode.READ_REQUEST,
                            request.handle,
                            BleAttErrorCode.INSUFFICIENT_AUTHOR
                        )
                    except HookReturnAccessDenied:
                        self.error(
                            BleAttOpcode.READ_REQUEST,
                            request.handle,
                            BleAttErrorCode.READ_NOT_PERMITTED
                        )
                    except HookReturnNotFound:
                        self.error(
                            BleAttOpcode.READ_REQUEST,
                            request.handle,
                            BleAttErrorCode.ATTRIBUTE_NOT_FOUND
                        )
                    except HookReturnGattError as gatt_error:
                        self.error(
                            gatt_error.request if gatt_error.request is not None else BleAttOpcode.READ_REQUEST,
                            gatt_error.handle if gatt_error.handle is not None else request.handle,
                            gatt_error.error if gatt_error.error is not None else BleAttErrorCode.ATTRIBUTE_NOT_FOUND
                        )
                else:
                    # Characteristic is not readable
                    self.error(
                        BleAttOpcode.READ_REQUEST,
                        request.handle,
                        BleAttErrorCode.READ_NOT_PERMITTED
                    )
            elif isinstance(attr, Characteristic):
                # Return characteristic value
                self.att.read_response(
                    attr.payload()
                )
            elif isinstance(attr, PrimaryService):
                # Return primary service value
                self.att.read_response(
                    attr.payload()
                )
            elif isinstance(attr, Descriptor):
                # Make sure the returned value matches the boundaries
                self.att.read_response(
                    attr.value[:local_mtu - 1]
                )

        except IndexError:
            self.error(
                BleAttOpcode.READ_REQUEST,
                request.handle,
                BleAttErrorCode.ATTRIBUTE_NOT_FOUND
            )

    @txlock
    def on_read_blob_request(self, request: GattReadBlobRequest):
        """Read blob request
        """
        if request.handle == 0:
            return self.error(
                BleAttOpcode.READ_BLOB_REQUEST,
                request.handle,
                BleAttErrorCode.INVALID_HANDLE
            )

        try:
            local_mtu = self.att.get_client_mtu()

            # Search attribute by handle and send response
            attr = self.server_model.find_object_by_handle(request.handle)

            if request.offset < len(attr.value):

                # If attribute is a characteristic, make sure it is readable
                # before returning a value.
                if isinstance(attr, CharacteristicValue):
                    try:
                        charac = self.server_model.find_object_by_handle(request.handle - 1)
                        service = self.server_model.find_service_by_characteristic_handle(charac.handle)

                        conn_handle = self.get_layer('l2cap').get_conn_handle()
                        if charac.check_security_property(ReadAccess, Authentication):
                            print("[i] authentication required for read access !")
                            if not self.get_layer('ll').state.is_authenticated(conn_handle):
                                self.error(
                                    BleAttOpcode.READ_BLOB_REQUEST,
                                    request.handle,
                                    BleAttErrorCode.INSUFFICIENT_AUTHENT
                                )
                                return
                        if charac.check_security_property(ReadAccess, Encryption):
                            print("[i] encryption required for read access !")
                            if not self.get_layer('ll').state.is_encrypted(conn_handle):
                                self.error(
                                    BleAttOpcode.READ_BLOB_REQUEST,
                                    request.handle,
                                    BleAttErrorCode.INSUFFICIENT_ENCRYPTION
                                )
                                return
                        if charac.check_security_property(ReadAccess, Authorization):
                            print("[i] authorization required for read access !")
                            # TODO: not supported for now
                            self.error(
                                BleAttOpcode.READ_BLOB_REQUEST,
                                request.handle,
                                BleAttErrorCode.INSUFFICIENT_AUTHOR
                            )
                            return
                        if not charac.readable():
                            self.error(
                                BleAttOpcode.READ_BLOB_REQUEST,
                                request.handle,
                                BleAttErrorCode.READ_NOT_PERMITTED
                            )
                            return

                        # Call our characteristic read hook
                        self.server_model.on_characteristic_read(
                            service,
                            charac,
                            request.offset,
                            local_mtu - 1
                        )

                        # Make sure the returned value matches the boundaries
                        value = charac.value[request.offset:request.offset + local_mtu - 1]

                        self.att.read_blob_response(
                            value
                        )

                    except HookReturnValue as force_value:
                        # Make sure the returned value matches the boundaries
                        value = force_value.value[:local_mtu - 1]

                        self.att.read_blob_response(
                            value
                        )
                    except HookReturnAuthentRequired:
                        self.error(
                            BleAttOpcode.READ_BLOB_REQUEST,
                            request.handle,
                            BleAttErrorCode.INSUFFICIENT_AUTHENT
                        )
                    except HookReturnAuthorRequired:
                        self.error(
                            BleAttOpcode.READ_REQUEST,
                            request.handle,
                            BleAttErrorCode.INSUFFICIENT_AUTHOR
                        )
                    except HookReturnAccessDenied:
                        self.error(
                            BleAttOpcode.READ_BLOB_REQUEST,
                            request.handle,
                            BleAttErrorCode.READ_NOT_PERMITTED
                        )
                    except HookReturnNotFound:
                        self.error(
                            BleAttOpcode.READ_BLOB_REQUEST,
                            request.handle,
                            BleAttErrorCode.ATTRIBUTE_NOT_FOUND
                        )
                    except HookReturnGattError as gatt_error:
                        self.error(
                            gatt_error.request if gatt_error.request is not None else BleAttOpcode.READ_REQUEST,
                            gatt_error.handle if gatt_error.handle is not None else request.handle,
                            gatt_error.error if gatt_error.error is not None else BleAttErrorCode.ATTRIBUTE_NOT_FOUND
                        )
                elif isinstance(attr, Descriptor):
                    # Valid offset, return data[offset:offset + MTU - 1]
                    self.att.read_blob_response(
                        attr.value[request.offset:request.offset + local_mtu - 1]
                    )
            elif request.offset == len(attr.value):
                # Special case: when offset == attribute length then return empty data
                self.att.read_blob_response(b'')
            else:
                # Invalid offset
                self.error(
                    BleAttOpcode.READ_BLOB_REQUEST,
                    request.handle,
                    BleAttErrorCode.INVALID_OFFSET
                )
        except IndexError:
            # Attribute not found
            self.error(
                BleAttOpcode.READ_BLOB_REQUEST,
                request.handle,
                BleAttErrorCode.ATTRIBUTE_NOT_FOUND
            )

    @txlock
    def on_write_request(self, request):
        """Write request for characteristic or descriptor value
        """
        if request.handle == 0:
            return self.error(
                BleAttOpcode.WRITE_REQUEST,
                request.handle,
                BleAttErrorCode.INVALID_HANDLE
            )

        try:
            # Retrieve attribute from model
            attr = self.server_model.find_object_by_handle(request.handle)
            if isinstance(attr, CharacteristicValue):
                # Check the corresponding characteristic is writeable
                charac = self.server_model.find_object_by_handle(request.handle - 1)

                conn_handle = self.get_layer('l2cap').get_conn_handle()
                if charac.check_security_property(WriteAccess, Authentication):
                    print("[i] authentication required for write access !")
                    if not self.get_layer('ll').state.is_authenticated(conn_handle):
                        self.error(
                            BleAttOpcode.WRITE_REQUEST,
                            request.handle,
                            BleAttErrorCode.INSUFFICIENT_AUTHENT
                        )
                        return
                if charac.check_security_property(WriteAccess, Encryption):
                    print("[i] encryption required for write access !")
                    if not self.get_layer('ll').state.is_encrypted(conn_handle):
                        self.error(
                            BleAttOpcode.WRITE_REQUEST,
                            request.handle,
                            BleAttErrorCode.INSUFFICIENT_ENCRYPTION
                        )
                        return
                if charac.check_security_property(WriteAccess, Authorization):
                    print("[i] authorization required for write access !")
                    # TODO: not supported for now
                    self.error(
                        BleAttOpcode.WRITE_REQUEST,
                        request.handle,
                        BleAttErrorCode.INSUFFICIENT_AUTHOR
                    )
                    return
                if charac.writeable():
                    # Retrieve corresponding service info
                    service = self.server_model.find_service_by_characteristic_handle(charac.handle)

                    try:
                        # Trigger our write hook
                        self.server_model.on_characteristic_write(
                            service,
                            charac,
                            0,
                            request.value,
                            False
                        )

                        # Update attribute value
                        attr.value = request.value
                        self.att.write_response()

                        # Trigger our written hook (after charac has been written)
                        self.server_model.on_characteristic_written(
                            service,
                            charac,
                            0,
                            attr.value,
                            True
                        )

                    except HookReturnValue as force_value:
                        # Make sure the returned value matches the boundaries
                        attr.value = force_value.value
                        self.att.write_response()

                        # Trigger our written hook (after charac has been written)
                        self.server_model.on_characteristic_written(
                            service,
                            charac,
                            0,
                            attr.value,
                            True
                        )

                    except HookReturnAuthentRequired:
                        self.error(
                            BleAttOpcode.WRITE_REQUEST,
                            request.handle,
                            BleAttErrorCode.INSUFFICIENT_AUTHENT
                        )
                    except HookReturnAuthorRequired:
                        self.error(
                            BleAttOpcode.READ_REQUEST,
                            request.handle,
                            BleAttErrorCode.INSUFFICIENT_AUTHOR
                        )
                    except HookReturnAccessDenied:
                        self.error(
                            BleAttOpcode.WRITE_REQUEST,
                            request.handle,
                            BleAttErrorCode.READ_NOT_PERMITTED
                        )
                    except HookReturnNotFound:
                        self.error(
                            BleAttOpcode.WRITE_REQUEST,
                            request.handle,
                            BleAttErrorCode.ATTRIBUTE_NOT_FOUND
                        )
                    except HookReturnGattError as gatt_error:
                        self.error(
                            gatt_error.request if gatt_error.request is not None else BleAttOpcode.READ_REQUEST,
                            gatt_error.handle if gatt_error.handle is not None else request.handle,
                            gatt_error.error if gatt_error.error is not None else BleAttErrorCode.ATTRIBUTE_NOT_FOUND
                        )
                else:
                    self.error(
                        BleAttOpcode.WRITE_REQUEST,
                        request.handle,
                        BleAttErrorCode.WRITE_NOT_PERMITTED
                    )
            elif isinstance(attr, ClientCharacteristicConfig):
                # Fixed length, make sure size <= 2.
                if len(request.value) <= 2:
                    attr.value = request.value + attr.value[len(request.value):]
                    self.att.write_response()

                    # Notify our model
                    if attr.config == 0x0001:
                        charac = attr.characteristic
                        service = self.server_model.find_service_by_characteristic_handle(charac.handle)

                        # Set characteristic notification callback
                        charac.set_notification_callback(self.notify)

                        self.server_model.on_characteristic_subscribed(
                            service,
                            charac,
                            notification=True
                        )
                    elif attr.config == 0x0002:
                        charac = attr.characteristic
                        service = self.server_model.find_service_by_characteristic_handle(charac.handle)

                        # Set characteristic indication callback
                        charac.set_indication_callback(self.indicate)

                        self.server_model.on_characteristic_subscribed(
                            service,
                            charac,
                            indication=True
                        )
                    elif attr.config == 0x0000:
                        charac = attr.characteristic
                        service = self.server_model.find_service_by_characteristic_handle(charac.handle)

                        # Unset characteristic indication and notification callbacks
                        charac.set_notification_callback(None)
                        charac.set_indication_callback(None)

                        if charac in self.__subscribed_characs:
                            self.__subscribed_characs.remove(charac)

                        # Notify model
                        self.server_model.on_characteristic_unsubscribed(
                            service,
                            charac
                        )
                else:
                    # Wrong length
                    self.error(
                        BleAttOpcode.WRITE_REQUEST,
                        request.handle,
                        BleAttErrorCode.INVALID_ATTR_VALUE_LENGTH
                    )
        except IndexError:
            self.error(
                BleAttOpcode.WRITE_REQUEST,
                request.handle,
                BleAttErrorCode.ATTRIBUTE_NOT_FOUND
            )

    @txlock
    def on_write_command(self, request):
        """Write command (without response)
        """
        if request.handle == 0:
            return self.error(
                BleAttOpcode.WRITE_COMMAND,
                request.handle,
                BleAttErrorCode.INVALID_HANDLE
            )

        try:
            # Retrieve attribute from model
            attr = self.server_model.find_object_by_handle(request.handle)
            if isinstance(attr, CharacteristicValue):
                # Check the corresponding characteristic is writeable
                charac = self.server_model.find_object_by_handle(request.handle - 1)

                conn_handle = self.get_layer('l2cap').get_conn_handle()
                if charac.check_security_property(WriteAccess, Authentication):
                    print("[i] authentication required for write access !")
                    if not self.get_layer('ll').state.is_authenticated(conn_handle):
                        self.error(
                            BleAttOpcode.WRITE_COMMAND,
                            request.handle,
                            BleAttErrorCode.INSUFFICIENT_AUTHENT
                        )
                        return
                if charac.check_security_property(WriteAccess, Encryption):
                    print("[i] encryption required for write access !")
                    if not self.get_layer('ll').state.is_encrypted(conn_handle):
                        self.error(
                            BleAttOpcode.WRITE_COMMAND,
                            request.handle,
                            BleAttErrorCode.INSUFFICIENT_ENCRYPTION
                        )
                        return
                if charac.check_security_property(WriteAccess, Authorization):
                    print("[i] authorization required for write access !")
                    # TODO: not supported for now
                    self.error(
                        BleAttOpcode.WRITE_COMMAND,
                        request.handle,
                        BleAttErrorCode.INSUFFICIENT_AUTHOR
                    )
                    return
                if charac.writeable():
                    # Retrieve corresponding service info
                    service = self.server_model.find_service_by_characteristic_handle(charac.handle)

                    try:
                        # Trigger our write hook
                        self.server_model.on_characteristic_write(
                            service,
                            charac,
                            0,
                            request.value,
                            True
                        )

                        # Update attribute value
                        attr.value = request.value

                        # Trigger our written hook (after charac has been written)
                        self.server_model.on_characteristic_written(
                            service,
                            charac,
                            0,
                            attr.value,
                            True
                        )

                    except HookReturnValue as force_value:
                        # Make sure the returned value matches the boundaries
                        attr.value = force_value.value

                        # Trigger our written hook (after charac has been written)
                        self.server_model.on_characteristic_written(
                            service,
                            charac,
                            0,
                            attr.value,
                            True
                        )

                    except HookReturnAuthentRequired:
                        self.error(
                            BleAttOpcode.WRITE_COMMAND,
                            request.handle,
                            BleAttErrorCode.INSUFFICIENT_AUTHENT
                        )
                    except HookReturnAuthorRequired:
                        self.error(
                            BleAttOpcode.WRITE_COMMAND,
                            request.handle,
                            BleAttErrorCode.INSUFFICIENT_AUTHOR
                        )
                    except HookReturnAccessDenied:
                        self.error(
                            BleAttOpcode.WRITE_COMMAND,
                            request.handle,
                            BleAttErrorCode.READ_NOT_PERMITTED
                        )
                    except HookReturnNotFound:
                        self.error(
                            BleAttOpcode.WRITE_COMMAND,
                            request.handle,
                            BleAttErrorCode.ATTRIBUTE_NOT_FOUND
                        )
                    except HookReturnGattError as gatt_error:
                        self.error(
                            gatt_error.request if gatt_error.request is not None else BleAttOpcode.READ_REQUEST,
                            gatt_error.handle if gatt_error.handle is not None else request.handle,
                            gatt_error.error if gatt_error.error is not None else BleAttErrorCode.ATTRIBUTE_NOT_FOUND
                        )
                else:
                    self.error(
                        BleAttOpcode.WRITE_COMMAND,
                        request.handle,
                        BleAttErrorCode.WRITE_NOT_PERMITTED
                    )

            elif isinstance(attr, ClientCharacteristicConfig):
                # Fixed length, make sure size <= 2.
                if len(request.value) <= 2 and request.value != attr.value:
                    attr.value = request.value + attr.value[len(request.value):]

                    # Notify our model
                    if attr.config == 0x0001:
                        charac = attr.characteristic
                        service = self.server_model.find_service_by_characteristic_handle(charac.handle)

                        # Set characteristic notification callback
                        charac.set_notification_callback(self.notify)
                        if charac not in self.__subscribed_characs:
                            self.__subscribed_characs.append(charac)

                        self.server_model.on_characteristic_subscribed(
                            service,
                            charac,
                            notification=True
                        )
                    elif attr.config == 0x0002:
                        charac = attr.characteristic
                        service = self.server_model.find_service_by_characteristic_handle(charac.handle)

                        # Set characteristic indication callback
                        charac.set_indication_callback(self.indicate)
                        if charac not in self.__subscribed_characs:
                            self.__subscribed_characs.append(charac)

                        self.server_model.on_characteristic_subscribed(
                            service,
                            charac,
                            indication=True
                        )
                    elif attr.config == 0x0000:
                        charac = attr.characteristic
                        service = self.server_model.find_service_by_characteristic_handle(charac.handle)

                        # Unset characteristic indication and notification callbacks
                        charac.set_notification_callback(None)
                        charac.set_indication_callback(None)

                        if charac in self.__subscribed_characs:
                            self.__subscribed_characs.remove(charac)

                        # Notify model
                        self.server_model.on_characteristic_unsubscribed(
                            service,
                            charac
                        )
                else:
                    # Wrong length
                    self.error(
                        BleAttOpcode.WRITE_COMMAND,
                        request.handle,
                        BleAttErrorCode.INVALID_ATTR_VALUE_LENGTH
                    )
        except IndexError:
            self.error(
                BleAttOpcode.WRITE_COMMAND,
                request.handle,
                BleAttErrorCode.ATTRIBUTE_NOT_FOUND
            )

    @txlock
    def on_prepare_write_request(self, request: GattPrepareWriteRequest):
        """Prepare write request
        """
        try:
            # Retrieve attribute from model
            attr = self.server_model.find_object_by_handle(request.handle)

            # Queue request
            if request.handle not in self.__write_queues:
                self.__write_queues[request.handle] = []
            self.__write_queues[request.handle].append(request)

            # Send response
            self.att.prepare_write_response(
                request.handle,
                request.offset,
                request.value
            )

        except IndexError:
            self.error(
                BleAttOpcode.PREPARE_WRITE_REQUEST,
                request.handle,
                BleAttErrorCode.INVALID_HANDLE
            )

    @txlock
    def on_execute_write_request(self, request: GattExecuteWriteRequest):
        """Execute write request
        """
        # Clear all prepared write queues
        if request.flags == 0:
            self.__write_queues = {}
            self.att.execute_write_response()
        elif request.flags == 1:
            # Apply write requests to items
            for handle in self.__write_queues:
                try:
                    # Retrieve attribute from model
                    attr = self.server_model.find_object_by_handle(handle)

                    if isinstance(attr, CharacteristicValue):
                        # apply each update
                        for write_req in self.__write_queues[handle]:
                            attr_value = attr.value
                            if write_req.offset > len(attr_value):
                                # Clear queues
                                self.__write_queues = {}

                                # Send error
                                self.error(
                                    BleAttOpcode.EXECUTE_WRITE_REQUEST,
                                    handle,
                                    BleAttErrorCode.INVALID_OFFSET
                                )

                                # Stop now
                                return
                            else:
                                if len(attr_value) >= (write_req.offset + len(write_req.value)):
                                    attr_value = attr_value[:write_req.offset] + write_req.value + attr_value[write_req.offset + len(write_req.value):]
                                else:
                                    attr_value = attr_value[:write_req.offset] + write_req.value
                                attr.value = attr_value
                    else:
                        # Nope, only characteristic values are supported
                        pass

                except IndexError:
                    # Clear write queues
                    self.__write_queues = {}

                    # Send error
                    self.error(
                        BleAttOpcode.EXECUTE_WRITE_REQUEST,
                        request.handle,
                        BleAttErrorCode.INVALID_HANDLE
                    )

                    # Done
                    return

            # Done !
            self.att.execute_write_response()
        else:
            # Unknown flag !
            pass

    @txlock
    def on_read_by_type_request(self, request: GattReadByTypeRequest):
        """Read attribute by type request
        """
        # Make sure the requested start handle is valid:
        # - start handle must not be 0
        # - start handle must be lower or equal to end handle
        if request.start == 0 or request.start > request.end:
            self.error(
                BleAttOpcode.READ_BY_TYPE_REQUEST,
                request.start,
                BleAttErrorCode.INVALID_HANDLE
            )
        else:
            # List attributes by type UUID, sorted by handles
            attrs = {}
            attrs_handles = []
            for attribute in self.server_model.attr_by_type_uuid(UUID(request.type), request.start, request.end):
                attrs[attribute.handle] = attribute
                attrs_handles.append(attribute.handle)
            attrs_handles.sort()

            # If we have at least one item to return
            if len(attrs_handles) > 0:

                # Get MTU
                mtu = self.att.get_client_mtu()

                # If client is looking for characteristic declaration,
                # we compute the correct item size and maximum number
                # of items we can put in a response PDU
                #
                # In the case of a characteristic declaration, UUID could
                # be 16-bit or 128-bit long so we shall only put in the same
                # answer characteristics with same size UUIDs.

                if UUID(request.type) == UUID(0x2803):

                    # Get item size (UUID size + 2)
                    uuid_size = len(attrs[attrs_handles[0]].uuid.packed)
                    item_size = uuid_size + 5
                    max_nb_items = int((mtu - 2) / item_size)

                    # Create our datalist
                    datalist = GattAttributeDataList(item_size)
                elif UUID(request.type) == UUID(0x2802):

                    # If client is looking for included services,
                    # we compute the correct item size and maximum number
                    # of items we can put in a response PDU
                    #
                    # In the case of an included service declaration, UUID could
                    # be 16-bit or 128-bit long so we shall only put in the same
                    # answer characteristics with same size UUIDs.

                    # Get item size
                    uuid_size = len(attrs[attrs_handles[0]].uuid.packed)
                    item_size = uuid_size + 6
                    max_nb_items = int((mtu - 2) / item_size)

                    # Create our datalist
                    datalist = GattAttributeDataList(item_size)
                else:
                    max_nb_items = 0
                    datalist = GattAttributeDataList(0)

                # Iterate over items while UUID size matches and data fits in MTU
                for i in range(max_nb_items):
                    if i < len(attrs_handles):
                        handle = attrs_handles[i]
                        attr_obj = attrs[handle]
                        if len(attr_obj.uuid.packed) == uuid_size:
                            if isinstance(attrs[handle], Characteristic):
                                datalist.append(
                                    GattAttributeValueItem(
                                        handle,
                                        pack(
                                            '<BH',
                                            attr_obj.properties,
                                            attr_obj.value_handle,
                                        ) + attr_obj.uuid.packed
                                    )
                                )
                            elif isinstance(attrs[handle], IncludeService):
                                if attrs[handle].service_uuid.type == UUID.TYPE_16:
                                    datalist.append(
                                        GattAttributeValueItem(
                                            handle,
                                            pack(
                                                '<HH',
                                                attr_obj.service_start_handle,
                                                attr_obj.service_end_handle,
                                            ) + attr_obj.service_uuid.packed
                                        )
                                    )
                                else:
                                    datalist.append(
                                        GattAttributeValueItem(
                                            handle,
                                            pack(
                                                '<HH',
                                                attr_obj.service_start_handle,
                                                attr_obj.service_end_handle,
                                            )
                                        )
                                    )
                    else:
                        break

                # Check that our result datalist does contain something
                if len(datalist) > 0:
                    # Once datalist created, send answer
                    datalist_raw = datalist.to_bytes()
                    self.att.read_by_type_response(item_size, datalist_raw)
                else:
                    # If not, send an error.
                    self.error(
                        BleAttOpcode.READ_BY_TYPE_REQUEST,
                        request.start,
                        BleAttErrorCode.ATTRIBUTE_NOT_FOUND
                    )
            else:
                self.error(
                   BleAttOpcode.READ_BY_TYPE_REQUEST,
                   request.start,
                   BleAttErrorCode.ATTRIBUTE_NOT_FOUND
                )

    @txlock
    def on_read_by_group_type_request(self, request: GattReadByGroupTypeRequest):
        """Read by group type request

        List attribute with given type UUID from `start` handle to ̀`end` handle.
        """
        # Make sure the requested start handle is valid:
        # - start handle must not be 0
        # - start handle must be lower or equal to end handle
        if request.start == 0 or request.start > request.end:
            self.error(
                BleAttOpcode.READ_BY_GROUP_TYPE_REQUEST,
                request.start,
                BleAttErrorCode.INVALID_HANDLE
            )
        elif UUID(request.type) not in SUPPORTED_GROUP_TYPES:
            # If request type UUID is unknown, return an error.
            self.error(
                BleAttOpcode.READ_BY_GROUP_TYPE_REQUEST,
                request.start,
                BleAttErrorCode.UNSUPPORTED_GROUP_TYPE
            )
        else:
            # List attributes by type UUID, sorted by handles
            attrs = {}
            attrs_handles = []
            for attribute in self.server_model.attr_by_type_uuid(UUID(request.type), request.start, request.end):
                attrs[attribute.handle] = attribute
                attrs_handles.append(attribute.handle)
            attrs_handles.sort()

            # If we have at least one item to return
            if len(attrs_handles) > 0:

                # Get MTU
                mtu = self.att.get_client_mtu()

                # Get item size (UUID size + 4)
                uuid_size = len(attrs[attrs_handles[0]].uuid.packed)
                item_size = uuid_size + 4
                max_nb_items = int((mtu - 2) / item_size)

                # Create our datalist
                datalist = GattAttributeDataList(item_size)

                # Iterate over items while UUID size matches and data fits in MTU
                for i in range(max_nb_items):
                    if i < len(attrs_handles):
                        handle = attrs_handles[i]
                        end_handle = attrs[handle].end_handle
                        attr_uuid = attrs[handle].uuid
                        if len(attr_uuid.packed) == uuid_size:
                            datalist.append(
                                GattGroupTypeItem(handle, end_handle, attr_uuid.packed)
                            )
                        else:
                            break
                    else:
                        break

                # Once datalist created, send answer
                datalist_raw = datalist.to_bytes()
                self.att.read_by_group_type_response(item_size, datalist_raw)
            else:
                self.error(
                   BleAttOpcode.READ_BY_GROUP_TYPE_REQUEST,
                   request.start,
                   BleAttErrorCode.ATTRIBUTE_NOT_FOUND
                )

    def on_terminated(self):
        """Connection has been terminated, remove characteristics subscriptions.
        """
        # Unsubscribe from everything and remove callbacks again
        for charac in self.__subscribed_characs:
            charac.set_notification_callback(None)
            charac.set_indication_callback(None)
        self.__subscribed_characs = []

    def pairing(self, pairing=None):
        pairing_parameters = Pairing()
        if pairing is None:
            pairing_parameters = self.server_model.get_pairing_parameters()
        elif isinstance(pairing, Pairing):
            pairing_parameters = pairing

        self.smp.request_pairing(pairing=pairing_parameters)

    @txlock
    def on_exch_mtu_request(self, request):
        """Received an MTU exchange request
        """
        # We do accept MTU exchange request, save server MTU and answer
        self.att.set_client_mtu(request.mtu)
        self.att.set_server_mtu(request.mtu)
        self.att.exch_mtu_response(self.att.get_server_mtu())

        # Notify our connector about this MTU update
        logging.debug("[gatt] notifying connector new client mtu: %d", request.mtu)
        self.att.notify_client_mtu()

    def on_exch_mtu_response(self, response):
        """ATT MTU exchange response

        :param GattExchangeMtuResponse response: response from remote device
        """
        self.on_gatt_message(response)

class GattClientServer(GattServer, GattClient):

    def __init__(self, parent=None, layer_name=None, options={}):
        super().__init__(parent=parent, layer_name=layer_name, options=options)



"""
Bluetooth Low Energy Tiny Stack - LLCAP
"""
from typing import List, Union

from scapy.packet import Packet, Raw
from scapy.layers.bluetooth import L2CAP_Hdr, ATT_Hdr
from scapy.layers.bluetooth4LE import BTLE_DATA

from .client import GattClient
from .server import GattServer

class Llcap:
    """L2CAP layer"""

    def __init__(self, gatt, conn_handle: int):
        """Initialize state."""
        self.__mtu = 23
        self.__conn_handle = conn_handle

        # Fragmentation
        self.__rx_fifo = b""
        self.__tx_fifo:List[Packet] = []
        self.__exp_length = 0

        # Configure GATT
        self.__gatt = gatt
        self.__gatt.set_l2cap(self)

    def on_pdu(self, packet: Packet, fragment: bool = False) -> List[Packet]:
        """Process an L2CAP packet."""
        if fragment:
            # Enqueue the whole L2CAP packet (including header)
            self.__rx_fifo += bytes(packet[L2CAP_Hdr])
        else:
            # Enqueue received bytes into our fifo
            l2cap_data = packet[L2CAP_Hdr]
            self.__rx_fifo = bytes(l2cap_data)

            # Retrieve expected length ( + header size)
            self.__exp_length = l2cap_data.len + 4

        # Check if we received a complete L2CAP frame
        if len(self.__rx_fifo) >= self.__exp_length:
            result = self.on_l2cap(L2CAP_Hdr(self.__rx_fifo[:self.__exp_length]))
            self.__rx_fifo = self.__rx_fifo[self.__exp_length:]
            self.__exp_length = 0
            return result

        # Nothing to do
        return []

    def on_l2cap(self, packet: Packet) -> list[Packet]:
        """Process complete L2CAP packet."""
        packets: list[Packet] = []

        # Check CID
        if packet.cid == 4:
            # Forward to GATT and convert response packets into L2CAP fragments
            for answer in self.__gatt.on_pdu(packet[ATT_Hdr]):
                packets.extend(self.__convert(answer))
        return packets

    def send_pdu(self, pdu: Packet):
        """Send a given ATT PDU to the host."""
        self.__tx_fifo.extend(self.__convert(pdu))

    def get_pdus(self) -> List[Packet]:
        """Retrieve PDUs to send from TX fifo."""
        pdus = self.__tx_fifo
        self.__tx_fifo = []
        return pdus

    def __convert(self, packet: Packet) -> list[Packet]:
        """Convert GATT packet into fragmented L2CAP packets if required."""
        # Create the answer L2CAP packet
        l2cap_pkt = L2CAP_Hdr(cid=4)/packet

        # Fragment it based on MTU
        l2cap_data = bytes(l2cap_pkt)
        nb_fragments = len(l2cap_data)//self.__mtu
        if nb_fragments*self.__mtu < len(l2cap_data):
            nb_fragments += 1

        if nb_fragments > 1:
            # Build BTLE_DATA packets
            packets = []
            for frag_id in range(nb_fragments):
                fragment = l2cap_data[frag_id*self.__mtu:(frag_id+1)*self.__mtu]
                packets.append(BTLE_DATA(LLID=0x02 if frag_id == 0 else 0x01)/Raw(fragment))
            return packets
        else:
            return [BTLE_DATA(LLID=0x02)/l2cap_pkt]

class LlcapClient(Llcap):
    """L2CAP layer with a GATT client sub-layer."""

    def __init__(self, conn_handle: int):
        """Initialize state."""
        self.__gatt_client = GattClient()
        super().__init__(self.__gatt_client, conn_handle)

    def get_gatt(self):
        return self.__gatt_client

class LlcapServer(Llcap):
    """L2CAP layer with a GATT server sub-layer."""

    def __init__(self, conn_handle: int):
        """Initialize state."""
        self.__gatt_server = GattServer()
        super().__init__(self.__gatt_server, conn_handle)

    def get_gatt(self):
        return self.__gatt_server




from whad.zigbee.stack.apl.zdo.object import ZDOObject
from whad.zigbee.stack.nwk.constants import NWKJoinMode
from whad.zigbee.stack.apl.constants import LogicalDeviceType

from whad.exceptions import RequiredImplementation

# Wrappers
from whad.zigbee.profile.network import Network
from whad.zigbee.profile.nodes import CoordinatorNode, EndDeviceNode, RouterNode

from time import sleep
from random import randint

import logging
logger = logging.getLogger(__name__)

class ZDONetworkManager(ZDOObject):
    """
    ZDO Device Object handling the network-related operations.
    """

    def configure_short_address(self, short_address):
        """
        Configure the short address to use.
        """
        phy_layer = self.zdo.manager.get_layer('phy')
        mac_layer = self.zdo.manager.get_layer('mac')
        nwk_layer = self.zdo.manager.get_layer('nwk')

        nwk_layer.database.set("nwkNetworkAddress", short_address)
        mac_layer.database.set("macShortAddress", short_address)
        phy_layer.set_short_address(short_address)

    def configure_extended_address(self, extended_address):
        """
        Configure the extended address to use.
        """
        phy_layer = self.zdo.manager.get_layer('phy')
        mac_layer = self.zdo.manager.get_layer('mac')
        nwk_layer = self.zdo.manager.get_layer('nwk')

        nwk_layer.database.set("nwkIeeeAddress", extended_address)
        mac_layer.database.set("macExtendedAddress", extended_address)
        phy_layer.set_extended_address(extended_address)

    def configure_extended_pan_id(self, extended_pan_id):
        """
        Configure the extended PanID of the network.
        """
        nwk_layer = self.zdo.manager.get_layer('nwk')

        nwk_layer.database.set("nwkExtendedPANID", extended_pan_id)

    def initialize(self, logical_type=LogicalDeviceType.END_DEVICE):
        """
        Initialize the network manager according to current configuration.
        """
        aps_layer = self.zdo.manager.get_layer('aps')

        # Network profile wrapper
        self.network = None
        # By default we are unauthorized on the network
        self.authorized = False

        # Get the logical type
        logical_type = self.zdo.configuration.get("configNodeDescriptor").logical_type

        # Check if we are an end device
        if logical_type == LogicalDeviceType.END_DEVICE:
            self.configure_extended_address(randint(0, 0xffffffffffffffff))
            self.configure_short_address(0xFFFF)
            self.configure_extended_pan_id(0x0000000000000000)
            aps_layer.database.set("apsDesignatedCoordinator", False)
            aps_layer.database.set("apsChannelMask", 0x7fff800)
            aps_layer.database.set("apsUseExtendedPANID", 0x0000000000000000)
            aps_layer.database.set("apsUseInsecureJoin", True)

        elif logical_type == LogicalDeviceType.COORDINATOR:
            self.configure_extended_address(randint(0, 0xffffffffffffffff))
            self.configure_short_address(0x0000)
            aps_layer.database.set("apsDesignatedCoordinator", True)
            aps_layer.database.set("apsUseExtendedPANID", randint(0, 0xffffffffffffffff))
            self.configure_extended_pan_id(0x0000000000000000)
            aps_layer.database.set("apsUseInsecureJoin", True)
            aps_layer.database.set(
                "apsTrustCenterAddress",
                self.zdo.nwk_management.get("nwkIeeeAddress")
            )

    def on_authorization(self):
        """
        Callback called when we are authorized.
        """
        # Mark the network as authorized.
        self.authorized = True
        # Indicate the device presence with a device announce.
        self.zdo.device_and_service_discovery.device_annce()

    def discover_networks(self):
        """
        Discover the surrounding networks.
        """
        phy_layer = self.zdo.manager.get_layer('phy')
        nwk_layer = self.zdo.manager.get_layer('nwk')
        nwk_management = nwk_layer.get_service('management')

        logger.info("[zdo_network_manager] Discovering networks.")
        networks = {}
        for network in nwk_management.network_discovery():
            # Create a wrapper for network
            networks[network.extended_pan_id] = Network(
                    network,
                    stack = phy_layer
            )
            # Loop over identified neighbors matching the current network
            for device_address, device in nwk_layer.database.get("nwkNeighborTable").table.items():
                if device.extended_pan_id == network.extended_pan_id:
                    if device.address == 0:
                        networks[network.extended_pan_id].coordinator = CoordinatorNode(
                            device.address,
                            extended_address=device.extended_address,
                            network=networks[network.extended_pan_id]
                        )
                    else:
                        networks[network.extended_pan_id].routers.append(
                            RouterNode(
                                device.address,
                                extended_address=device.extended_address,
                                network=networks[network.extended_pan_id]
                            )
                        )
        return networks.values()

    def join(self, network, force: bool = False):
        """
        Join a specific network.
        """
        nwk_layer = self.zdo.manager.get_layer('nwk')
        nwk_management = nwk_layer.get_service('management')
        aps_layer = self.zdo.manager.get_layer('aps')
        aps_management = aps_layer.get_service('management')

        nwkExtendedPANID = nwk_management.get("nwkExtendedPANID")
        apsDesignatedCoordinator = aps_management.get("apsDesignatedCoordinator")
        apsUseInsecureJoin = aps_management.get("apsUseInsecureJoin")
        apsChannelMask = aps_management.get("apsChannelMask")

        aps_layer.database.set("apsUseExtendedPANID", network.extended_pan_id)
        apsUseExtendedPANID = aps_management.get("apsUseExtendedPANID")
        logger.info("[zdo_network_manager] Joining specific network: %s.", repr(network))

        join_success = nwk_management.join(
            apsUseExtendedPANID,
            association_type=NWKJoinMode.NEW_JOIN,
            scan_channels=apsChannelMask,
            force=force
        )
        if not join_success:
            logger.info("[zdo_network_manager] failure during network join, exiting.")
            return False

        logger.debug("[zdo_network_manager] successfully joined network.")
        self.network = network
        return True



    def rejoin(self, network):
        """
        Rejoin a specific network.
        """
        nwk_layer = self.zdo.manager.get_layer('nwk')
        nwk_management = nwk_layer.get_service('management')
        aps_layer = self.zdo.manager.get_layer('aps')
        aps_management = aps_layer.get_service('management')

        nwkExtendedPANID = nwk_management.get("nwkExtendedPANID")
        apsDesignatedCoordinator = aps_management.get("apsDesignatedCoordinator")
        apsUseInsecureJoin = aps_management.get("apsUseInsecureJoin")
        apsChannelMask = aps_management.get("apsChannelMask")

        aps_layer.database.set("apsUseExtendedPANID", network.extended_pan_id)
        apsUseExtendedPANID = aps_management.get("apsUseExtendedPANID")
        logger.info("[zdo_network_manager] Rejoining specific network: %s.", repr(network))

        join_success = nwk_management.join(
            apsUseExtendedPANID,
            association_type=NWKJoinMode.REJOIN,
            scan_channels=apsChannelMask,
            security_enable=True
        )
        if not join_success:
            logger.info("[zdo_network_manager] failure during network rejoin, exiting.")
            return False

        self.network = network
        self.on_authorization()
        return True

    def leave(self):
        """
        Leave the network if we are currently associated.
        """
        nwk_layer = self.zdo.manager.get_layer('nwk')
        nwk_management = nwk_layer.get_service('management')

        return nwk_management.leave()

    def startup(self):
        nwk_layer = self.zdo.manager.get_layer('nwk')
        nwk_management = nwk_layer.get_service('management')
        aps_layer = self.zdo.manager.get_layer('aps')
        aps_management = aps_layer.get_service('management')
        phy_layer = self.zdo.manager.get_layer('phy')


        nwkExtendedPANID = nwk_management.get("nwkExtendedPANID")
        apsDesignatedCoordinator = aps_management.get("apsDesignatedCoordinator")
        apsUseExtendedPANID = aps_management.get("apsUseExtendedPANID")
        apsUseChannel = aps_management.get("apsUseChannel")

        apsUseInsecureJoin = aps_management.get("apsUseInsecureJoin")
        apsChannelMask = aps_management.get("apsChannelMask")
        if nwkExtendedPANID != 0:
            # we are already connected :) don't do anything
            return True

        if apsDesignatedCoordinator:
            # We are a coordinator not connected, start NLME-NETWORK-FORMATION
            if apsUseExtendedPANID != 0:
                nwk_management.set("nwkExtendedPANID", apsUseExtendedPANID)

            nwk_management.network_formation(
                pan_id=None,
                channel=apsUseChannel,
                scan_channels=apsChannelMask,
            )

            self.network = Network(
                nwk_management.get("nwkOwnNetwork"), phy_layer
            )
            self.authorized = True
            return True
        else:
            # We are a router or an end device, attempt to join or rejoin a network
            if apsUseExtendedPANID != 0:
                rejoin_success = nwk_management.join(
                    apsUseExtendedPANID,
                    association_type=NWKJoinMode.REJOIN,
                    scan_channels=apsChannelMask
                )
                if rejoin_success:
                    return True

                # We failed to rejoin, try to join
                selected_zigbee_network = None
                for attempt in range(self.zdo.configuration.get("configNWKScanAttempts")):
                    zigbee_networks = nwk_management.network_discovery(scan_channels=apsChannelMask)
                    for zigbee_network in zigbee_networks:
                        if zigbee_network.extended_pan_id == apsUseExtendedPANID:
                            selected_zigbee_network = zigbee_network
                    if selected_zigbee_network is not None:
                        break
                    else:
                        sleep(
                                self.zdo.configuration.get("configNWKTimeBetweenScans") *
                                phy_layer.symbol_duration * 2
                        )
                if selected_zigbee_network is None:
                    logger.info("[zdo_network_manager] target network not found, exiting.")
                    return False

                join_success = nwk_management.join(
                    apsUseExtendedPANID,
                    association_type=NWKJoinMode.NEW_JOIN,
                    scan_channels=apsChannelMask
                )
                if not join_success:
                    logger.info("[zdo_network_manager] failure during network join, exiting.")
                self.network = Network(
                        selected_zigbee_network,
                        stack = phy_layer
                )
                return join_success

            else:
                # join the best available network
                selected_zigbee_network = None
                for attempt in range(self.zdo.configuration.get("configNWKScanAttempts")):
                    zigbee_networks = nwk_management.network_discovery(scan_channels=apsChannelMask)
                    for zigbee_network in zigbee_networks:
                        if zigbee_network.joining_permit:
                            selected_zigbee_network = zigbee_network
                    if selected_zigbee_network is not None:
                        break
                    else:
                        sleep(
                                self.zdo.configuration.get("configNWKTimeBetweenScans") *
                                phy_layer.symbol_duration * 2
                        )
                if selected_zigbee_network is None:
                    logger.info("[zdo_network_manager] no target network found, exiting.")
                    return False

                join_success = nwk_management.join(
                    selected_zigbee_network.extended_pan_id,
                    association_type=NWKJoinMode.NEW_JOIN,
                    scan_channels=apsChannelMask
                )
                if not join_success:
                    logger.info("[zdo_network_manager] failure during network join, exiting.")
                self.network = Network(
                        selected_zigbee_network,
                        stack = phy_layer
                )
                return join_success



"""Bluetooth Low Energy connection hijacker connector.
"""
from whad.ble.connector.base import BLE
from whad.ble.connector.central import Central
from whad.ble.connector.peripheral import Peripheral
from whad.ble.exceptions import ConnectionLostException
from whad.exceptions import UnsupportedCapability
from whad.helpers import message_filter
from whad.hub.events import ConnectionEvt
from whad.hub.ble import Hijacked

class Hijacker(BLE):
    """Bluetooth Low Energy hijacking connector.
    """

    def __init__(self, device, connection=None):
        super().__init__(device)
        self.__connection = connection
        self.__hijack_master = False
        self.__hijack_slave = False
        self.__status = False
        self.__exception = None
        # Check if device accepts hijacking
        if not self.can_hijack_slave() and not self.can_hijack_master():
            raise UnsupportedCapability("Hijack")

    @property
    def central(self) -> Central:
        """Return the hijacker central.
        """
        available_actions = self.available_actions(Central)
        if len(available_actions) == 1:
            return available_actions[0]
        return None

    @property
    def peripheral(self):
        """Return the hijacker peripheral.
        """
        available_actions = self.available_actions(Peripheral)
        if len(available_actions) == 1:
            return available_actions[0]
        return None

    def available_actions(self, action_filter=None):
        """Determine the possible actions once a connection has been hijacked.
        """
        actions = []
        if self.__status:
            # It should be replaced by arguments in function calls, building a\
            # pseudo packet seems dirty
            if self.__hijack_master:
                pseudo_connection = ConnectionEvt()
                pseudo_connection.conn_handle = 0
                pseudo_connection.initiator = b"\x00\x00\x00\x00\x00\x00"
                pseudo_connection.init_addr_type = 0
                pseudo_connection.advertiser = b"\x00\x00\x00\x00\x00\x00"
                pseudo_connection.adv_addr_type = 0
                actions.append(Central(self.device, existing_connection=pseudo_connection))
            if self.__hijack_slave:
                pseudo_connection = ConnectionEvt()
                pseudo_connection.conn_handle = 1
                pseudo_connection.initiator = b"\x00\x00\x00\x00\x00\x00"
                pseudo_connection.init_addr_type = 0
                pseudo_connection.advertiser = b"\x00\x00\x00\x00\x00\x00"
                pseudo_connection.adv_addr_type = 0

                actions.append(
                    Peripheral(
                        self.device,
                        existing_connection = pseudo_connection
                    )
                )
        return [action for action in actions if action_filter is None or isinstance(action, filter)]


    def hijack(self, master = True, slave = False):
        """
        Hijack master, slave, or both
        """
        if self.__connection is not None:
            if master and slave:
                self.__hijack_master = master
                self.__hijack_slave = slave
                self.hijack_both(self.__connection.access_address)
            elif master:
                self.__hijack_master = master
                self.hijack_master(self.__connection.access_address)

            elif slave:
                self.__hijack_slave = slave
                self.hijack_slave(self.__connection.access_address)

            message = self.wait_for_message(keep=message_filter(Hijacked))
            self.__status = message.success
            return message.success

        # Connection failed
        raise self.__exception

    def on_desynchronized(self, access_address=None):
        """Desynchronization callback
        """
        if access_address == self.__connection.access_address:
            self.__exception = ConnectionLostException(self.__connection)
            self.__connection = None



"""
ReadBlob procedure
"""
from typing import List

from scapy.packet import Packet
from scapy.layers.bluetooth import ATT_Error_Response, ATT_Read_Blob_Request, ATT_Read_Blob_Response

from whad.ble.stack.att.constants import BleAttErrorCode, BleAttOpcode

from .attribute import find_attr_by_handle
from .procedure import BleClientProcedure, BleServerProcedure, UnexpectedProcError

class ServerReadBlobProcedure(BleServerProcedure):
    """ATT server ReadBlob procedure."""

    # ATT Operation code
    OPCODE = BleAttOpcode.READ_BLOB_REQUEST

    def __init__(self, attributes: list, mtu: int):
        """Initialize our ServerReadBlob procedure."""
        super().__init__(attributes, mtu)

    @classmethod
    def trigger(cls, request) -> bool:
        """Determine if the procedure is triggered."""
        return ATT_Read_Blob_Request in request

    def process_request(self, request: Packet) -> list[Packet]:
        """React only on ReadBlobRequest."""
        if ATT_Read_Blob_Request not in request:
            self.set_state(self.states.ERROR)
            raise UnexpectedProcError()

        request = request[ATT_Read_Blob_Request]

        try:
            # Query the requested attribute
            attr = find_attr_by_handle(self.attributes, request.gatt_handle)

            # Make sure we can read this attribute
            if not attr.readable():
                self.set_state(self.states.DONE)
                return self.att_error_response(request.gatt_handle, BleAttErrorCode.READ_NOT_PERMITTED)

            # Check offset
            if request.offset >= len(attr.value):
                self.set_state(self.states.DONE)
                return self.att_error_response(request.gatt_handle, BleAttErrorCode.INVALID_OFFSET)

            # Return value
            return [ ATT_Read_Blob_Response(value=attr.value[request.offset:self.mtu-1]) ]
        except IndexError:
            self.set_state(self.states.DONE)
            return self.att_error_response(request.gatt_handle, BleAttErrorCode.ATTRIBUTE_NOT_FOUND)

class ClientReadBlobProcedure(BleClientProcedure):
    """GATT Client ReadBlob Procedure."""

    # ATT Operation code
    OPCODE = BleAttOpcode.READ_BLOB_RESPONSE

    def __init__(self, handle: int, offset: int):
        """Initialize a ReadBlob procedure."""
        super().__init__([
            ATT_Read_Blob_Request(
                gatt_handle=handle,
                offset=offset
            )
        ])

    def process_request(self, request: Packet) -> List[Packet]:
        """Process incoming PDU."""
        if ATT_Error_Response in request:
            self.set_result(request[ATT_Error_Response])
            self.set_state(self.states.ERROR)
        elif ATT_Read_Blob_Response in request:
            response = request[ATT_Read_Blob_Response]
            self.set_result(response.value)
            self.set_state(self.states.DONE)
        else:
            self.set_result(None)
            self.set_state(self.states.ERROR)
        return []

